"""
Test inference for a simple data-generating model for data derived from a Normal distribution:

    X_1, ..., X_n ~ Normal(mu, sigma)

The marginal posteriors should be:

    mu | X = x ~ T-Dist(mean(x), std(x) / sqrt(n))
    sigma^2 | X = x ~ Inv-Gamma((n-1) / 2, (n-1)s^2 / 2)

With expectation:

    E[mu | X = x] = mean(x)
    E[sigma^2 | X = x]

and variance:

    Var(mu | X = x) =
"""

import jax.random as jr
from jaxtyping import Array, Scalar

import bayinx as byx
from bayinx import define
from bayinx.dists import Normal
from bayinx.flows import DiagAffine
from bayinx.nodes import Continuous, Observed


# Define model
class SimpleNormalModel(byx.Model, init=False):
    mu: Continuous[Scalar] = define(shape = ())
    sigma: Continuous[Scalar] = define(shape = (), lower = 0.0)

    x: Observed[Array] = define(shape = 'n_obs')

    def model(self, target):
        self.x << Normal(self.mu, self.sigma)

        return target

# Simulate sample
n_obs = 100
x: Array = jr.normal(jr.key(0), (n_obs, )) * 5 + 10

def test_inference():
    # Define posterior
    posterior = byx.Posterior(SimpleNormalModel, n_obs = n_obs, x = x)
    posterior.configure([DiagAffine()])
    posterior.fit()

    # Get posterior
    mu_draws = posterior.sample('mu', int(1e6))
    #sigma_draws = posterior.sample('sigma', int(1e6))

    # Confirm approximation is accurate
    # mu | X ~ t_n-1 (mu = xbar, sigma = s / sqrt(n)) ==> E[mu | X] = xbar
    # sigma^2 | X ~ inv-gamma (alpha = (n-1) / 2, beta = (n-1)s^2 / 2) ==> E[sigma^2 | X] =
    assert abs(mu_draws.mean() - x.mean()) < 0.01
    assert abs(mu_draws.var(ddof = 1) - x.var(ddof = 1) / n_obs * (n_obs - 1) / (n_obs - 3)) < 0.1 # this is not passing for low 'n_obs'



from typing import Tuple

import equinox as eqx
import jax.lax as lax
import jax.numpy as jnp
import jax.random as jr
import jax.scipy.special as jsp
import jax.tree as jt
from jaxtyping import Array, ArrayLike, PRNGKeyArray, Real, Scalar

from bayinx.core.distribution import Distribution
from bayinx.core.node import Node
from bayinx.nodes import Observed

PI = 3.141592653589793


def prob(
    x: Real[ArrayLike, "..."],
    mu: Real[ArrayLike, "..."],
    sigma: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The probability density function (PDF) for a Normal distribution.

    # Parameters
    - `x`: Where to evaluate the PDF.
    - `mu`: The mean.
    - `sigma`: The standard deviation.

    # Returns
    The PDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `mu`, and `sigma`.
    """
    # Cast to Array
    x, mu, sigma = jnp.asarray(x), jnp.asarray(mu), jnp.asarray(sigma)

    return lax.exp(-0.5 * lax.square((x - mu) / sigma)) / (sigma * lax.sqrt(2.0 * PI))


def logprob(
    x: Real[ArrayLike, "..."],
    mu: Real[ArrayLike, "..."],
    sigma: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the probability density function (log PDF) for a Normal distribution.

    # Parameters
    - `x`: Where to evaluate the log PDF.
    - `mu`: The mean.
    - `sigma`: The standard deviation.

    # Returns
    The log PDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `mu`, and `sigma`.
    """
    # Cast to Array
    x, mu, sigma = jnp.asarray(x), jnp.asarray(mu), jnp.asarray(sigma)

    return -lax.log(lax.sqrt(2.0 * PI)) - lax.log(sigma) - 0.5 * lax.square((x - mu) / sigma)


def cdf(
    x: Real[ArrayLike, "..."],
    mu: Real[ArrayLike, "..."],
    sigma: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The cumulative density function (CDF) for a Normal distribution.

    # Parameters
    - `x`: Where to evaluate the CDF.
    - `mu`: The mean.
    - `sigma`: The standard deviation.

    # Returns
    The CDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `mu`, and `sigma`.
    """
    # Cast to Array
    x, mu, sigma = jnp.asarray(x), jnp.asarray(mu), jnp.asarray(sigma)

    return jsp.ndtr((x - mu) / sigma)


def logcdf(
    x: Real[ArrayLike, "..."],
    mu: Real[ArrayLike, "..."],
    sigma: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the cumulative density function (log CDF) for a Normal distribution.

    # Parameters
    - `x`: Where to evaluate the log CDF.
    - `mu`: The mean.
    - `sigma`: The standard deviation.

    # Returns
    The log CDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `mu`, and `sigma`.
    """
    # Cast to Array
    x, mu, sigma = jnp.asarray(x), jnp.asarray(mu), jnp.asarray(sigma)

    return jsp.log_ndtr((x - mu) / sigma)


def ccdf(
    x: Real[ArrayLike, "..."],
    mu: Real[ArrayLike, "..."],
    sigma: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The complementary cumulative density function (cCDF) for a Normal distribution.

    # Parameters
    - `x`: Where to evaluate the cCDF.
    - `mu`: The mean.
    - `sigma`: The standard deviation.

    # Returns
    The cCDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `mu`, and `sigma`.
    """
    # Cast to Array
    x, mu, sigma = jnp.asarray(x), jnp.asarray(mu), jnp.asarray(sigma)

    return jsp.ndtr((mu - x) / sigma)


def logccdf(
    x: Real[ArrayLike, "..."],
    mu: Real[ArrayLike, "..."],
    sigma: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the complementary cumulative density function (log cCDF) for a Normal distribution.

    # Parameters
    - `x`: Where to evaluate the log cCDF.
    - `mu`: The mean.
    - `sigma`: The standard deviation.

    # Returns
    The log cCDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `mu`, and `sigma`.
    """
    # Cast to Array
    x, mu, sigma = jnp.asarray(x), jnp.asarray(mu), jnp.asarray(sigma)

    return jsp.log_ndtr((mu - x) / sigma)


class Normal(Distribution):
    """
    A normal distribution.

    # Attributes
    - `mu`: The mean/location parameter.
    - `sigma`: The standard-deviation/scale parameter.
    """

    mu: Node[Real[Array, "..."]]
    sigma: Node[Real[Array, "..."]]


    def __init__(
        self,
        mu: Real[ArrayLike, "..."] | Node[Real[Array, "..."]],
        sigma: Real[ArrayLike, "..."] | Node[Real[Array, "..."]]
    ):
        # Initialize mean/location parameter (mu)
        if isinstance(mu, Node):
            if isinstance(mu.obj, ArrayLike):
                self.mu = mu # type: ignore
        else:
            self.mu = Observed(jnp.asarray(mu))

        # Initialize dispersion/scale parameter (sigma)
        if isinstance(sigma, Node):
            if isinstance(sigma.obj, ArrayLike):
                self.sigma = sigma # type: ignore
        else:
            self.sigma = Observed(jnp.asarray(sigma))


    def logprob(self, node: Node) -> Scalar:
        obj, mu, sigma = (
            node.obj,
            self.mu.obj,
            self.sigma.obj
        )

        # Filter out irrelevant values
        obj, _ = eqx.partition(obj, node._filter_spec)

        # Helper function for the single-leaf log-probability evaluation
        def leaf_logprob(x: Real[ArrayLike, "..."]) -> Scalar:
            return logprob(x, mu, sigma).sum()

        # Compute log probabilities across leaves
        eval_obj = jt.map(leaf_logprob, obj)

        # Compute total sum
        total = jt.reduce_associative(lambda x,y: x + y, eval_obj, identity=0.0)

        return jnp.asarray(total)

    def sample(self, shape: int | Tuple[int, ...], key: PRNGKeyArray = jr.key(0)):
        # Coerce to tuple
        if isinstance(shape, int):
            shape = (shape, )

        return jr.normal(key, shape) * self.sigma.obj + self.mu.obj



from typing import Self, Tuple

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jax.flatten_util import ravel_pytree
from jaxtyping import Array, PRNGKeyArray, PyTree, Scalar

from bayinx.core.variational import M, Variational
from bayinx.dists import normal


class Standard(Variational[M]):
    """
    A standard normal approximation to a posterior distribution.

    # Attributes
    - `dim`: The dimension of the support.
    """

    def __init__(self, model: M):
        """
        Constructs a standard normal approximation to a posterior distribution.

        # Parameters
        - `model`: A probabilistic `Model` object.
        """
        # Partition model
        params, self._static = eqx.partition(model, model.filter_spec)

        # Flatten params component
        params, self._unflatten = ravel_pytree(params)

        # Store dimension of parameter space
        self.dim = jnp.size(params)


    @eqx.filter_jit
    def sample(self, n: int, key: PRNGKeyArray = jr.PRNGKey(0)) -> Array:
        # Sample variational draws
        draws: Array = jr.normal(key=key, shape=(n, self.dim))

        # Shape checks
        assert len(draws.shape) == 2

        return draws

    @eqx.filter_jit
    def eval(self, draws: Array) -> Array:
        return normal.logprob(
            x=draws,
            mu=jnp.array(0.0),
            sigma=jnp.array(1.0),
        ).sum(axis=1)

    @property
    def filter_spec(self):
        filter_spec = jtu.tree_map(lambda _: False, self)

        return filter_spec

    @eqx.filter_jit
    def elbo(self, n: int, batch_size: int, key: PRNGKeyArray) -> Scalar:
        dyn, static = eqx.partition(self, self.filter_spec)

        @eqx.filter_jit
        def elbo(dyn: Self, n: int, key: PRNGKeyArray) -> Scalar:
            vari = eqx.combine(dyn, static)

            # Sample draws from variational distribution
            draws: Array = vari.sample(n, key)

            # Evaluate posterior density for each draw
            posterior_evals: Array = vari.eval_model(draws)

            # Evaluate variational density for each draw
            variational_evals: Array = vari.eval(draws)

            # Evaluate ELBO
            return jnp.mean(posterior_evals - variational_evals)

        return elbo(dyn, n, key)

    @eqx.filter_jit
    def elbo_grad(self, n: int, batch_size: int, key: PRNGKeyArray) -> Self:
        raise RuntimeError("Do not use the 'elbo_grad' method for a Standard variational approximation. It has no variational parameters.")
        return self

    def elbo_and_grad(self, n: int, batch_size: int, key: PRNGKeyArray) -> Tuple[Scalar, PyTree]:
        """
        Evaluate the ELBO and its gradient.
        """
        raise RuntimeError("Do not use the 'elbo_and_grad' method for a Standard variational approximation. It has no variational parameters.")
        return self.elbo(n, key), self



from abc import abstractmethod
from dataclasses import field, fields
from typing import (
    Dict,
    Optional,
    Self,
    Tuple,
    Type,
    get_origin,
    get_type_hints,
)

import equinox as eqx
import jax.numpy as jnp
import jax.tree as jt
from jaxtyping import PyTree, Scalar

from bayinx.constraints import Identity, Interval, Lower, Upper
from bayinx.core.context import _model_context, model_context
from bayinx.core.node import Node
from bayinx.core.types import HasConstraint
from bayinx.core.utils import _extract_shape_params, _resolve_shape_spec
from bayinx.nodes import Continuous, Observed, Stochastic


def define(
    shape: Optional[int | str | Tuple[int | str, ...]] = None,
    init: Optional[PyTree] = None,
    lower: Optional[float] = None,
    upper: Optional[float] = None
):
    """
    Define a stochastic node.
    """
    metadata: Dict = {}

    if shape is not None:
        metadata["shape"] = shape
    if init is not None:
        metadata["init"] = init

    match (lower, upper):
        case (float() | int(), None):
            metadata["constraint"] = Lower(lower) # type: ignore
        case (None, float() | int()):
            metadata["constraint"] = Upper(upper) # type: ignore
        case (float() | int(), float() | int()):
            metadata["constraint"] = Interval(lower, upper) # type: ignore
        case (None, None):
            metadata["constraint"] = Identity()
        case (_):
            raise TypeError("TODO.")

    return field(metadata=metadata)


class Model(eqx.Module):
    """
    A base class used to define probabilistic models.
    """

    def __init_subclass__(cls, **kwargs):
        # Consume 'init' argument before passing it up to Equinox
        kwargs.pop('init', None)
        super().__init_subclass__(**kwargs)

    def __init__(self, **kwargs):
        cls = self.__class__

        # Grab initialized parameters
        init_params: set[str] = {f.name for f in fields(cls) if f.name in kwargs.keys()}

        # Grab shape parameters from model definition
        shape_params: set[str] = set()
        for node_defn in fields(cls):
            if (shape_spec := node_defn.metadata.get("shape")) is not None:
                shape_params = shape_params | _extract_shape_params(shape_spec)

        # Check all shape parameters are passed as arguments
        if not shape_params.issubset(kwargs.keys()):
            missing_params = shape_params - kwargs.keys()
            raise TypeError(
                f"Following shape parameters were not specified during model initialization: {", ".join(missing_params)}"
            )

        # Define all initialized dimensions
        shape_values: dict = {
            shape_param: kwargs[shape_param]
            for shape_param in shape_params
        }

        # Grab node types
        node_types: dict[str, Type[Node]] = {k: get_origin(v) for k, v in get_type_hints(cls).items()}


        # Auto-initialize parameters based on field metadata and type annotations
        for node_defn in fields(cls):
            # Grab node type
            node_type = node_types[node_defn.name]

            # Grab shape information if available
            shape_spec: str | None = node_defn.metadata.get("shape")
            shape = _resolve_shape_spec(shape_spec, shape_values)

            # Construct object
            if node_defn.name in init_params: # Initialized in model construction
                obj = kwargs[node_defn.name]
            elif "init" in node_defn.metadata: # Initialized in model definition
                obj = node_defn.metadata["init"]
            elif shape is not None: # Shape defined in model definition
                obj = jnp.zeros(shape) # TODO: will change later for discrete objects
            else:
                raise ValueError(f"Node '{node_defn.name}' not initialized or defined.")

            # Check shape
            if shape is not None and jnp.shape(obj) != shape:
                raise ValueError(f"Expected shape {shape} for {node_defn.name} but got {jnp.shape(obj)}.")

            if issubclass(node_type, Stochastic):
                if node_type == Continuous:
                    setattr(
                        self,
                        node_defn.name,
                        Continuous(obj, node_defn.metadata["constraint"]),
                    )
                else:
                    TypeError(f"{node_type.__name__} is not implemented yet")
            elif issubclass(node_type, Observed):
                setattr(self, node_defn.name, Observed(obj))
            else:
                raise TypeError(f"{node_defn.name} node is neither Stochastic nor Observed but {node_type.__name__}.")


    @property
    def filter_spec(self) -> Self:
        """
        Generates a filter specification to subset stochastic elements of the model.
        """
        # Generate empty specification
        filter_spec: Self = jt.map(lambda _: False, self)

        for f in fields(self): # type: ignore
            # Extract attribute
            node: Node = getattr(self, f.name)

            # Check if attribute is stochastic
            if isinstance(node, Stochastic):
                # Update model's filter specification at node
                filter_spec: Self = eqx.tree_at(
                    lambda model: getattr(model, f.name),
                    filter_spec,
                    replace=node.filter_spec
                )

        return filter_spec

    def filter_for(self, node_type: Type[Stochastic]) -> Self:
        """
        Generates a filter specification to subset stochastic elements of a certain type of the model.
        """
        # Generate empty specification
        filter_spec: Self = jt.map(lambda _: False, self)

        for f in fields(self): # type: ignore
            # Extract node
            node: Node = getattr(self, f.name)

            if isinstance(node, node_type):
                # Update model's filter specification for node
                filter_spec: Self = eqx.tree_at(
                    lambda model: getattr(model, f.name),
                    filter_spec,
                    replace=node.filter_spec,
                )

        return filter_spec


    def constrain(self, jacobian: bool = True) -> Tuple[Self, Scalar]:
        """
        Constrain nodes to the appropriate domain.

        # Returns
        A tuple containing the constrained `Model` object and the log-Jacobian adjustment.
        """
        model: Self = self
        target: Scalar = jnp.array(0.0)

        for f in fields(self): # type: ignore
            # Extract attribute
            node = getattr(self, f.name)

            # Check if node has a constraint
            if isinstance(node, HasConstraint):
                # Apply constraint
                obj, laj = node._constraint.constrain(node.obj, node._filter_spec)

                # Update values with constrained counterpart
                model = eqx.tree_at(
                    where=lambda model: getattr(model, f.name).obj,
                    pytree=model,
                    replace=obj,
                )

                # Adjust posterior density
                if jacobian:
                    target += laj

        return model, target

    @abstractmethod
    def model(self, target: Scalar) -> Scalar:
        pass

    @eqx.filter_jit
    def __call__(self) -> Scalar:
        with model_context():
            # Constrain the model and accumulate Jacobian adjustments
            self, target = self.constrain()

            # Accumulate manual increments
            target += self.model(jnp.array(0.0))

            # Accumulate implicit increments
            target += _model_context.target.value

            # Return the accumulated target
            return target



from typing import Callable, Dict, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Array, Float, PRNGKeyArray, Scalar

from bayinx.core.flow import FlowLayer, FlowSpec


def _h(x: Array) -> Array:
    return jnp.tanh(x)

def _dh(x: Array) -> Array:
    return 1.0 - jnp.tanh(x)**2


class PlanarLayer(FlowLayer):
    """
    A Planar flow.

    # Attributes
    - `params`: The parameters of the Planar flow (w, u_hat, b).
    - `constraints`: The constraining transformations for the parameters.
    - `static`: Whether the flow layer is frozen.
    """

    params: Dict[str, Array]
    constraints: Dict[str, Callable[[Array], Array]]
    static: bool

    def __init__(self, dim: int, key: PRNGKeyArray = jr.key(0)):
        """
        Initializes a Planar flow.

        # Parameters
        - `dim`: The dimension of the parameter space.
        """
        self.static = False
        # Split key
        k1, k2, k3 = jr.split(key, 3)

        # Initialize parameters
        self.params = {
            "w": jr.normal(k1, (dim,)) / dim**0.5,
            "u_hat": jr.normal(k2, (dim,)) / dim**0.5,
            "b": jr.normal(k3, ()) / dim**0.5,
        }

        self.constraints = {}

    def __forward(self, draw: Float[Array, " n_dim"]) -> Float[Array, " n_dim"]:
        params = self.transform_params()

        # Extract parameters
        w: Float[Array, " n_dim"] = params["w"]
        u: Float[Array, " n_dim"] = params["u"]
        b: Scalar = params["b"]

        # Compute inner term
        a = draw.dot(w) + b

        # Compute nonlinear stretch
        h = _h(a)

        # Compute forward transformation
        draw = draw + u * h

        return draw

    @eqx.filter_jit
    def forward(self, draws: Float[Array, "n_draws n_dim"]) -> Float[Array, "n_draws n_dim"]:
        f = jax.vmap(self.__forward, 0)
        return f(draws)

    def __adjust(self, draw: Float[Array, "n_dim"]) -> Scalar: # noqa
        params = self.transform_params()

        # Extract parameters
        w: Float[Array, " n_dim"] = params["w"]
        u: Float[Array, " n_dim"] = params["u"]
        b: Scalar = params["b"]

        # Compute inner term
        a = draw.dot(w) + b

        # Compute log-Jacobian adjustment
        lja: Scalar = jnp.log(1.0 + u.dot(_dh(a) * w))

        assert lja.shape == ()
        return lja

    @eqx.filter_jit
    def adjust(self, draws: Float[Array, "n_draws n_dim"]) -> Float[Array, "n_draws n_dim"]:
        f = jax.vmap(self.__adjust, 0)
        return f(draws)

    def __forward_and_adjust(self, draw: Float[Array, " n_dim"]) -> Tuple[Float[Array, " n_dim"], Scalar]:
        params = self.transform_params()

        # Extract parameters
        w: Float[Array, " n_dim"] = params["w"]
        u: Float[Array, " n_dim"] = params["u"]
        b: Scalar = params["b"]

        # Compute inner term
        a = draw.dot(w) + b

        # Compute forward transformation
        draw = draw + u * _h(a)

        # Compute log-Jacobian adjustment
        lja: Scalar = jnp.log(1.0 + u.dot(_dh(a) * w))

        assert len(draw.shape) == 1
        assert lja.shape == ()

        return draw, lja

    @eqx.filter_jit
    def forward_and_adjust(self, draws: Float[Array, "n_draws n_dim"]) -> Tuple[Float[Array, "n_draws n_dim"], Scalar]:
        f = jax.vmap(self.__forward_and_adjust, 0)
        return f(draws)


    def transform_params(self) -> Dict[str, Array]:
        """
        Applies the affine constraint to u_hat to compute the constrained parameter u,
        and returns all parameters.

        # Returns
        A dictionary of parameters including the computed, constrained 'u'.
        """
        constrained_params = self.constrain_params()

        # Extract parameters
        w: Array = constrained_params["w"]
        u_hat: Array = constrained_params["u_hat"]

        # Compute the constrained u
        u = u_hat + (-1.0 - jnp.dot(w, u_hat)) / (jnp.sum(w**2) + 1e-6) * w
        constrained_params["u"] = u

        return constrained_params

class Planar(FlowSpec):
    """
    A specification for the Planar flow.
    """
    def __init__(self):
        pass

    def construct(self, dim: int) -> PlanarLayer:
        return PlanarLayer(dim)



from typing import Callable, Optional, Self, Tuple

import equinox as eqx
import jax.flatten_util as jfu
import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jax.lax import scan
from jaxtyping import Array, PRNGKeyArray, Scalar

from bayinx.core.flow import FlowLayer
from bayinx.core.variational import M, Variational


class NormalizingFlow(Variational[M]):
    """
    An ordered collection of diffeomorphisms that map a base distribution to a
    normalized approximation of a posterior distribution.

    # Attributes
    - `dim`: The dimension of the support.
    - `base`: A base variational distribution.
    - `flows`: An ordered collection of continuously parameterized diffeomorphisms.
    """
    flows: list[FlowLayer]
    base: Variational[M]

    def __init__(
        self,
        base: Variational[M],
        flows: list[FlowLayer],
        model: Optional[M] = None,
        _static: Optional[M] = None,
        _unflatten: Optional[Callable[[Array], M]] = None
    ):
        """
        Constructs an unoptimized normalizing flow posterior approximation.

        # Parameters
        - `base`: The base variational distribution.
        - `flows`: A list of flows.
        - `model`: A probabilistic `Model` object.
        """
        if model is not None:
            # Partition model
            params, self._static = eqx.partition(model, model.filter_spec)

            # Flatten params component
            _, self._unflatten = jfu.ravel_pytree(params)
        elif _static is not None and _unflatten is not None:
            self._static = _static
            self._unflatten = _unflatten
        else:
            raise ValueError("Either 'model' or '_static' and '_unflatten' must be specified.")

        self.dim = base.dim
        self.base = base
        self.flows = flows

    @property
    def filter_spec(self) -> Self:
        # Generate empty specification
        filter_spec: Self = jtu.tree_map(lambda _: False, self)

        # Specify variational parameters based on each flow's filter spec.
        filter_spec: Self = eqx.tree_at(
            lambda vari: vari.flows,
            filter_spec,
            replace=[flow.filter_spec for flow in self.flows],
        )

        return filter_spec

    @eqx.filter_jit
    def sample(self, n: int, key: PRNGKeyArray = jr.PRNGKey(0)) -> Array:
        # Sample from the base distribution
        draws: Array = self.base.sample(n, key)

        assert len(draws.shape) == 2

        # Apply forward transformations
        for map in self.flows:
            draws = map.forward(draws)

        assert len(draws.shape) == 2

        return draws

    @eqx.filter_jit
    def eval(self, draws: Array) -> Array:
        raise RuntimeError("Evaluating the variational density for a normalizing flow requires an analytic inverse to exist, which many useful flows do not have. Therefore, do not use this method.")
        return jnp.full(draws.shape[0], jnp.nan)

    @eqx.filter_jit
    def __eval(self, draws: Array) -> Tuple[Array, Array]:
        """
        Evaluate the posterior and variational densities together at the
        transformed `draws` to avoid extra compute.

        # Parameters
        - `draws`: Draws from the base variational distribution.

        # Returns
        The posterior and variational densities as JAX Arrays.
        """
        # Evaluate base density
        variational_evals: Array = self.base.eval(draws)

        # Shape checks
        assert len(variational_evals.shape) == 1
        assert len(draws.shape) == 2

        for map in self.flows:
            # Apply transformation
            draws, ljas = map.forward_and_adjust(draws)
            assert len(draws.shape) == 2
            assert len(ljas.shape) == 1

            # Adjust variational density
            variational_evals = variational_evals - ljas

        # Evaluate posterior at final variational draws
        posterior_evals = self.eval_model(draws)

        # Shape checks
        assert len(posterior_evals.shape) == 1
        assert len(variational_evals.shape) == 1
        assert posterior_evals.shape == variational_evals.shape

        return posterior_evals, variational_evals

    @eqx.filter_jit
    def elbo(self, n: int, batch_size: int, key: PRNGKeyArray = jr.PRNGKey(0)) -> Scalar:
        dyn, static = eqx.partition(self, self.filter_spec)

        def elbo(dyn: Self, n: int, key: PRNGKeyArray) -> Scalar:
            self = eqx.combine(dyn, static)

            # Split keys
            keys = jr.split(key, n // batch_size)

            # Split ELBO calculation into batches
            def batched_elbo(carry: None, key: PRNGKeyArray) -> Tuple[None, Array]:
                # Draw from variational distribution
                draws: Array = self.base.sample(batch_size, key)

                # Evaluate posterior and variational densities
                batched_post_evals, batched_vari_evals = self.__eval(draws)

                # Compute ELBO estimate
                batched_elbo_evals: Array = batched_post_evals - batched_vari_evals

                return None, batched_elbo_evals

            elbo_evals = scan(
                batched_elbo,
                init=None,
                xs=keys,
                length=n // batch_size
            )[1]

            # Compute average of ELBO estimates
            return jnp.mean(elbo_evals)

        return elbo(dyn, n, key)

    @eqx.filter_jit
    def elbo_grad(self, n: int, batch_size: int, key: PRNGKeyArray) -> Self:
        dyn, static = eqx.partition(self, self.filter_spec)

        # Define ELBO function
        def elbo(dyn: Self, n: int, key: PRNGKeyArray) -> Scalar:
            self = eqx.combine(dyn, static)

            # Split key
            keys = jr.split(key, n // batch_size)

            # Split ELBO calculation into batches
            def batched_elbo(carry: None, key: PRNGKeyArray) -> Tuple[None, Array]:
                # Draw from variational distribution
                draws: Array = self.base.sample(batch_size, key)

                # Evaluate posterior and variational densities
                batched_post_evals, batched_vari_evals = self.__eval(draws)

                # Compute ELBO estimate
                batched_elbo_evals: Array = batched_post_evals - batched_vari_evals

                return None, batched_elbo_evals

            elbo_evals = scan(
                batched_elbo,
                init=None,
                xs=keys,
                length=n // batch_size
            )[1]

            # Compute average of ELBO estimates
            return jnp.mean(elbo_evals)

        # Map to its gradient
        elbo_grad: Callable[
            [Self, int, PRNGKeyArray], Self
        ] = eqx.filter_grad(elbo)

        return elbo_grad(dyn, n, key)

    @eqx.filter_jit
    def elbo_and_grad(self, n: int, batch_size: int, key: PRNGKeyArray) -> Tuple[Scalar, Self]:
        dyn, static = eqx.partition(self, self.filter_spec)

        def elbo(dyn: Self, n: int, key: PRNGKeyArray) -> Scalar:
            self = eqx.combine(dyn, static)

            # Split keys
            keys = jr.split(key, n // batch_size)

            # Split ELBO calculation into batches
            def batched_elbo(carry: None, key: PRNGKeyArray) -> Tuple[None, Array]:
                # Draw from variational distribution
                draws: Array = self.base.sample(batch_size, key)

                # Evaluate posterior and variational densities
                batched_post_evals, batched_vari_evals = self.__eval(draws)

                # Compute ELBO estimate
                batched_elbo_evals: Array = batched_post_evals - batched_vari_evals

                return None, batched_elbo_evals

            elbo_evals = scan(
                batched_elbo,
                init=None,
                xs=keys,
                length=n // batch_size
            )[1]

            # Compute average of ELBO estimates
            return jnp.mean(elbo_evals)

        # Map to its value & gradient
        elbo_and_grad: Callable[
            [Self, int, PRNGKeyArray], Tuple[Scalar, Self]
        ] = eqx.filter_value_and_grad(elbo)

        return elbo_and_grad(dyn, n, key)



# pyright: reportUnusedExpression=false

import jax.numpy as jnp
from jaxtyping import Array

from bayinx import Model, define
from bayinx.dists import Normal
from bayinx.nodes import Continuous, Observed


class SimpleDynamic(Model, init=False):
    mu: Continuous[Array] = define(shape = ())
    y: Observed[Array] = define()


    def model(self, target):
        self.y << Normal(self.mu, 1.0)

        return target


def test_simple():
    y: Array = jnp.array([-1.0, 0.0, 1.0])

    # Construct model
    model = SimpleDynamic(y = y)
    assert all(model.y.obj == jnp.array([-1.0, 0.0, 1.0]))


class ShapedDynamic(Model, init=False):
    mus: Continuous[Array] = define(shape = ('k'))
    y: Observed[Array] = define(shape = ('n', 'k'))

    def model(self, target):
        self.y << Normal(self.mus, 1.0)

        return target

def test_shaped():
    y: Array = jnp.tile(jnp.array([-1.0, 0.0, 1.0]), (10, 1))
    #bad_y: Array = jnp.tile(jnp.array([-1.0, 1.0]), (10, 1))

    model = ShapedDynamic(k = 3, n = 10, y = y)
    assert model.mus.obj.shape == (3,)
    assert model.y.obj.shape == (10,3)

class PackedShapeDynamic(Model, init=False):
    x: Continuous[Array] = define(shape = 'shape')
    y: Continuous[Array] = define(shape = ('m', 'n', 'shape'))

    def model(self, target):
        self.x << Normal(0.0, 1.0)

        return target

def test_packedshape():
    model = PackedShapeDynamic(shape = (3,2,1), m = 5, n = 4)
    assert model.x.obj.shape == (3,2,1)
    assert model.y.obj.shape == (5,4,3,2,1)




from functools import partial
from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, Type

import equinox as eqx
import jax
import jax.random as jr
from jax.lax import scan
from jaxtyping import Array, PRNGKeyArray

from bayinx.core.flow import FlowSpec
from bayinx.core.node import Node
from bayinx.core.variational import M
from bayinx.vi.normalizing_flow import NormalizingFlow
from bayinx.vi.standard import Standard


class Posterior(Generic[M]):
    """
    The posterior distribution for a model.

    # Attributes
    - `vari`: The variational approximation of the posterior.
    - `config` The configuration for the posterior.
    """
    vari: NormalizingFlow[M]
    config: Dict[str, Any]

    def __init__(self, model_def: Type[M], **kwargs: Any):
        # (hopefully) omit intermediate model construction through jit
        @eqx.filter_jit
        def construct_base(model_def):
            # Construct model
            model = model_def(**kwargs)

            return Standard(model)

        # Construct standard normal base distribution
        self.vari = construct_base(model_def)

        # Include default attributes
        self.config = {
            "learning_rate": 0.1 / self.vari.dim**0.5,
            "tolerance": 1e-4,
            "grad_draws": 4,
            "batch_size": 1
        }


    def configure(
        self,
        flowspecs: Optional[List[FlowSpec]] = None,
        learning_rate: Optional[float] = None,
        tolerance: Optional[float] = None,
        grad_draws: Optional[int] = None,
        batch_size: Optional[int] = None
    ):
        """
        Configure the variational approximation.

        # Parameters
        - `flowspecs`: The specification for a sequence of flows.
        """
        # Append new NF architecture
        if flowspecs is not None:
            # Initialize NF architecture
            flows = [
                flowspec.construct(self.vari.dim) for flowspec in flowspecs
            ]

            if isinstance(self.vari, Standard):
                # Construct new normalizing flow
                self.vari = NormalizingFlow(
                    base = self.vari,
                    flows = flows,
                    _static = self.vari._static,
                    _unflatten = self.vari._unflatten
                )
            elif isinstance(self.vari, NormalizingFlow):
                # Freeze current flows
                for flow in self.vari.flows:
                    object.__setattr__(flow, 'static', True) # kind of illegal but I need to avoid copies

                # Append new flows
                self.vari.flows.extend(flows)

        # Include other settings
        if learning_rate is not None:
            self.config["learning_rate"] = learning_rate
        if tolerance is not None:
            self.config["tolerance"] = tolerance
        if grad_draws is not None:
            self.config["grad_draws"] = grad_draws
        if batch_size is not None:
            self.config["batch_size"] = batch_size


    def fit(
        self,
        max_iters: int = 50_000,
        learning_rate: Optional[float] = None,
        tolerance: Optional[float] = None,
        grad_draws: Optional[int] = None,
        batch_size: Optional[int] = None,
        key: PRNGKeyArray = jr.key(0),
    ):
        # Include settings
        if learning_rate is not None:
            self.config["learning_rate"] = learning_rate
        if tolerance is not None:
            self.config["tolerance"] = tolerance
        if grad_draws is not None:
            self.config["grad_draws"] = grad_draws
        if batch_size is not None:
            self.config["batch_size"] = batch_size

        if isinstance(self.vari, Standard):
            # Construct default sequence of optimization
            print("TODO")
        else:
            # Optimize variational approximation with user-specified flows
            self.vari = self.vari.fit(
                max_iters,
                self.config["learning_rate"],
                self.config["tolerance"],
                self.config["grad_draws"],
                self.config["batch_size"],
                key
            )

    def sample(
        self,
        node: str,
        n_draws: int,
        batch_size: Optional[int] = None,
        key: PRNGKeyArray = jr.key(0)
    ) -> Array:
        """
        Sample a node from the posterior distribution.

        # Parameters
        - `node`: The name of the node.
        - `n_draws`: The number of draws from the posterior.
        - `batch_size`: The number of draws for the full model ever initialized in memory at once.
        - `key`: The PRNG key.
        """
        if batch_size is None:
            batch_size = n_draws

        # Split keys
        keys = jr.split(key, n_draws // batch_size)

        @partial(jax.vmap, in_axes = 0)
        def reconstruct_and_subset(draw: Array):
            model = self.vari.reconstruct_model(draw).constrain()[0]

            return getattr(model, node).obj

        def batched_sample(carry: None, key: PRNGKeyArray):
            # Sample draws
            draws = self.vari.sample(batch_size, key)

            return None, reconstruct_and_subset(draws)

        posterior_draws = scan(
            batched_sample,
            init=None,
            xs=keys,
            length=n_draws // batch_size
        )[1].squeeze()

        return posterior_draws


    def predictive(
        self,
        func: Callable[[M, PRNGKeyArray], Node[Array] | Array],
        n_draws: int,
        batch_size: Optional[int] = None,
        key: PRNGKeyArray = jr.key(0)
    ) -> Array:
        """

        """
        if batch_size is None:
            batch_size = n_draws

        # Split keys
        keys = jr.split(key, n_draws // batch_size)

        @partial(jax.vmap, in_axes = (0, 0))
        def reconstruct_and_predict(draw: Array, key: PRNGKeyArray) -> Array:
            model = self.vari.reconstruct_model(draw).constrain()[0]

            # Compute predictive
            obj = func(model, key)

            # Coerce from Node if needed
            if isinstance(obj, Node):
                obj: Array = obj.obj # type: ignore

            return obj

        def batched_sample(carry: None, key: PRNGKeyArray) -> Tuple[None, Array]:
            # Sample draws
            draws = self.vari.sample(batch_size, key)

            # Generate additional keys for each draw
            keys = jr.split(key, batch_size)

            return None, reconstruct_and_predict(draws, keys)

        posterior_draws: Array = scan(
            batched_sample,
            init=None,
            xs=keys,
            length=n_draws // batch_size
        )[1].squeeze()

        return posterior_draws



from typing import Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
from jax.lax import scan
from jaxtyping import Array, Float, PRNGKeyArray, Scalar

from bayinx.core.flow import FlowLayer, FlowSpec


def _mat_op(
    carry: Float[Array, " rank"],
    vars: Tuple[
        Float[Array, " rank"],
        Float[Array, " rank"],
        Scalar,
        Scalar,
        Scalar
    ]
) -> Tuple[Float[Array, " rank"], Scalar]:
    """
    The implicit matrix operations compute:

        x = Lz + b = (D + W)z + b = Dz + (UV^T * M)z + b

    Where W = UV^T * M is the low-rank representation of the strictly lower triangular matrix W,
        parameterized by U and V, and an implicit mask M that zeros out upper triangular elements.

    In the implementation only U, V, diag(D), and b are needed to complete the matrix operations.
    """
    # Unpack state
    U_r, V_r, z_i, d_i, b_i = vars
    h_i = carry

    # Compute partial product
    h_next = h_i + z_i * V_r

    # Compute i-th element of transformed draw:
    x_i = b_i + z_i * d_i + U_r.dot(h_i)

    return h_next, x_i

class LowRankAffineLayer(FlowLayer):
    """
    A low-rank affine flow.

    # Attributes
    - `params`: A dictionary containing the shift and low-rank representation of the scale parameters.
    - `constraints`: A dictionary of constraining transformations.
    - `static`: Whether the flow layer is frozen (parameters are not subject to further optimization).
    - `rank`: Rank of the scale transformation.
    """

    rank: int

    def __init__(self, dim: int, rank: int, key: PRNGKeyArray = jr.key(0)):
        """
        Initializes a low-rank affine flow.

        # Parameters
        - `dim`: The dimension of the parameter space.
        - `rank`: The rank of the (implicit) scale matrix.
        """
        self.static = False
        self.rank = rank

        # Split key
        k1, k2, k3, k4 = jr.split(key, 4)

        # Initialize parameters
        self.params = {
            "shift": jr.normal(k1, (dim, )) / dim**0.5,
            "diag_scale": jr.normal(k2, (dim, )) / dim**0.5,
            "offdiag_scale": (
                jr.normal(k3, (dim, rank)) / dim**0.5,
                jr.normal(k4, (dim, rank)) / dim**0.5
            )
        }

        # Define constraints
        self.constraints = {"diag_scale": jnp.exp}


    def __forward(self, draw: Float[Array, " n_dim"]) -> Float[Array, " n_dim"]:
        params = self.transform_params()

        # Extract parameters
        shift: Array = params["shift"]
        diag: Array = params["diag_scale"]
        U, V = params["offdiag_scale"]

        # Compute forward transformation
        _, draw = scan(
            f=_mat_op,
            init=jnp.zeros((self.rank, )),
            xs= (U, V, draw, diag, shift)
        )

        return draw

    @eqx.filter_jit
    def forward(self, draws: Float[Array, "n_draws n_dim"]) -> Float[Array, "n_draws n_dim"]:
        f = jax.vmap(self.__forward, 0)
        return f(draws)

    def __adjust(self, draw: Float[Array, " n_dim"]) -> Scalar:
        params = self.transform_params()

        diag: Array = params["diag_scale"]

        # Compute log-Jacobian adjustment
        lja: Scalar = jnp.log(diag).sum()

        assert lja.shape == ()

        return lja

    @eqx.filter_jit
    def adjust(self, draws: Float[Array, "n_draws n_dim"]) -> Float[Array, "n_draws n_dim"]:
        f = jax.vmap(self.__adjust, 0)
        return f(draws)

    def __forward_and_adjust(self, draw: Float[Array, " n_dim"]) -> Tuple[Float[Array, " n_dim"], Scalar]:
        params = self.transform_params()

        assert len(draw.shape) == 1

        # Extract parameters
        shift: Array = params["shift"]
        diag: Array = params["diag_scale"]
        U, V = params["offdiag_scale"]

        # Compute log-Jacobian adjustment
        lja: Scalar = jnp.log(diag).sum()

        # Compute forward transformation
        _, draw = scan(
            f=_mat_op,
            init=jnp.zeros((self.rank, )),
            xs= (U, V, draw, diag, shift)
        )

        assert len(draw.shape) == 1
        assert lja.shape == ()

        return draw, lja

    @eqx.filter_jit
    def forward_and_adjust(self, draws: Float[Array, "n_draws n_dim"]) -> Tuple[Float[Array, "n_draws n_dim"], Scalar]:
        f = jax.vmap(self.__forward_and_adjust, 0)
        return f(draws)

class LowRankAffine(FlowSpec):
    rank: int

    def __init__(self, rank: int):
        self.rank = rank

    def construct(self, dim: int) -> LowRankAffineLayer:
        if (self.rank > (dim - 1)/2):
            raise ValueError(f"Rank {self.rank} is large, consider using a full affine flow instead.")

        return LowRankAffineLayer(dim, self.rank)



from abc import abstractmethod
from functools import partial
from typing import Callable, Generic, Self, Tuple, TypeVar

import equinox as eqx
import jax
import jax.lax as lax
import jax.numpy as jnp
import jax.random as jr
import optax as opx
from jaxtyping import Array, PRNGKeyArray, PyTree, Scalar
from optax import GradientTransformation, OptState

from bayinx.core.model import Model

M = TypeVar("M", bound=Model)

class Variational(eqx.Module, Generic[M]):
    """
    An abstract base class used to define variational methods.

    # Attributes
    - `dim`: The dimension of the support.
    - `_unflatten`: A function to transform draws from the variational distribution back to a `Model`.
    - `_static`: The static component of a partitioned `Model` used to initialize the `Variational` object.
    """
    dim: int
    _unflatten: Callable[[Array], M]
    _static: M

    @property
    @abstractmethod
    def filter_spec(self) -> Self:
        """
        Filter specification for dynamic and static components of the
        `Variational` object.
        """
        pass

    @abstractmethod
    def sample(self, n: int, key: PRNGKeyArray = jr.PRNGKey(0)) -> Array:
        """
        Sample from the variational distribution.
        """
        pass

    @abstractmethod
    def eval(self, draws: Array) -> Array:
        """
        Evaluate the variational distribution at `draws`.
        """
        pass

    @abstractmethod
    def elbo(self, n: int, batch_size: int, key: PRNGKeyArray) -> Array:
        """
        Evaluate the ELBO.
        """
        pass

    @abstractmethod
    def elbo_grad(self, n: int, batch_size: int, key: PRNGKeyArray) -> M:
        """
        Evaluate the gradient of the ELBO.
        """
        pass

    @abstractmethod
    def elbo_and_grad(self, n: int, batch_size: int, key: PRNGKeyArray) -> Tuple[Scalar, PyTree]:
        """
        Evaluate the ELBO and its gradient.
        """
        pass

    @eqx.filter_jit
    def reconstruct_model(self, draw: Array) -> M:
        # Unflatten variational draw
        model: M = self._unflatten(draw)

        # Combine with static components
        model: M = eqx.combine(model, self._static)

        return model

    @eqx.filter_jit
    @partial(jax.vmap, in_axes=(None, 0))
    def eval_model(self, draws: Array) -> Array:
        """
        Reconstruct models from variational draws and evaluate their posterior.

        # Parameters
        - `draws`: A set of variational draws.
        """
        # Unflatten variational draw
        model: M = self.reconstruct_model(draws)

        # Evaluate posterior
        return model()

    @eqx.filter_jit
    def fit(
        self,
        max_iters: int,
        learning_rate: float,
        tolerance: float,
        grad_draws: int,
        batch_size: int,
        key: PRNGKeyArray = jr.key(0),
    ) -> Self:
        """
        Optimize the variational distribution.
        """
        # Partition variational
        dyn, static = eqx.partition(self, self.filter_spec)

        # Construct scheduler
        schedule = opx.warmup_cosine_decay_schedule(
            1e-8,
            learning_rate,
            int(max_iters / 10),
            max_iters - int(max_iters / 10)
        )

        # Initialize optimizer
        optim: GradientTransformation = opx.chain(
            opx.scale(-1.0), opx.adam(schedule) # replace learning_rate with scheduler
        )
        opt_state: OptState = optim.init(dyn)

        def condition(state: Tuple[Self, OptState, Scalar, PRNGKeyArray]):
            # Unpack iteration state
            dyn, opt_state, i, key = state

            return i < max_iters

        def body(state: Tuple[Self, OptState, Scalar, PRNGKeyArray]):
            # Unpack iteration state
            dyn, opt_state, i, key = state

            # Update iteration
            i = i + 1

            # Update PRNG key
            key, _ = jr.split(key)

            # Reconstruct variational
            vari: Self = eqx.combine(dyn, static)

            # Compute gradient of the ELBO for update
            update: M = vari.elbo_grad(grad_draws, batch_size, key)

            # Transform update through optimizer
            update, opt_state = optim.update( # type: ignore
                update, opt_state, eqx.filter(dyn, dyn.filter_spec) # type: ignore
            )

            # Update variational distribution
            dyn: Self = eqx.apply_updates(dyn, update)

            return dyn, opt_state, i, key

        # Run optimization loop
        dyn = lax.while_loop(
            cond_fun=condition,
            body_fun=body,
            init_val=(dyn, opt_state, jnp.array(0, jnp.uint32), key),
        )[0]

        # Return optimized variational
        return eqx.combine(dyn, static)



import jax.random as jr
from jaxtyping import Array, Scalar

import bayinx as byx
from bayinx import define
from bayinx.dists import Binomial
from bayinx.flows import DiagAffine
from bayinx.nodes import Continuous, Observed


# Define model
class SimpleBinomialModel(byx.Model, init=False):
    p: Continuous[Scalar] = define(shape = (), lower = 0, upper = 1)

    x: Observed[Array] = define(shape = 'n_obs')
    n: Observed[Array] = define(shape = ())

    def model(self, target):
        self.x << Binomial(self.n, self.p)

        return target

# Simulate sample
n_obs = 100
n = 10
x: Array = jr.binomial(jr.key(0), n, 0.5, (n_obs, ))

def test_inference():
    # Define posterior
    posterior = byx.Posterior(
        SimpleBinomialModel,
        n_obs = n_obs,
        n = n,
        x = x
    )
    posterior.configure([DiagAffine()])
    posterior.fit()

    # Get posterior
    p_draws = posterior.sample('p', int(1e6))
    #sigma_draws = posterior.sample('sigma', int(1e6))

    # Confirm approximation is accurate
    # p | X ~ beta(alpha = x + 1, beta = n + 1 - x) ==> E[p | X] = p
    # sigma^2 | X ~ inv-gamma (alpha = (n-1) / 2, beta = (n-1)s^2 / 2) ==> E[sigma^2 | X] =
    assert abs(p_draws.mean() - x.mean()) < 0.01
    assert abs(p_draws.var(ddof = 1) - x.var(ddof = 1) / n_obs * (n_obs - 1) / (n_obs - 3)) < 0.1 # this is not passing for low 'n_obs'



# pyright: reportUnusedExpression=false
from typing import Dict, List

import jax.numpy as jnp
from jaxtyping import Array

from bayinx import Model, define
from bayinx.dists import Normal
from bayinx.nodes import Continuous, Observed


# Define model
class MyModel(Model, init=False):
    x: Continuous[Array] = define() # shows up in __init__ method, typed as 'Array'
    y: Observed[Array] = define() # shows up in __init__ method, typed as 'Array'
    misc_1: Observed[str] = define(init = "hi")
    misc_2: Observed[List[str]] = define(init = ["ciao", "goodbye"])
    misc_3: Observed[Dict[str, str]] = define(init = {'key': 'value'})

    def model(self, target):
        self.y << Normal(self.x, 1.0)

        return target


def test_init():
    """
    Test aspects of initialization.
    """
    model = MyModel(x = jnp.array(0.0), y = jnp.array([0.0, 0.0, 0.0]))

    # Test filter specification
    assert model.filter_spec.x.obj # Stochastic nodes should be included
    assert not model.filter_spec.y.obj # Observed nodes should be excluded
    assert not model.filter_spec.misc_1.obj # Regardless of type
    assert not model.filter_spec.misc_2.obj[0] # But should be PyTrees internally

    # Test (log unnormalized) posterior evaluation
    assert abs(model() - jnp.array(-2.756815599614018453)) < 1e-6 # 3 * log P_Z(0)

def test_subset():
    """
    Test hard subsetting.
    """
    model = MyModel(x = jnp.array(0.0), y = jnp.array([0.0, 0.0, 0.0]))

    assert model.misc_2[0].obj == "ciao"
    assert model.misc_3['key'].obj == "value"



from typing import Tuple

import equinox as eqx
import jax.lax as lax
import jax.numpy as jnp
import jax.random as jr
import jax.scipy.special as jsp
import jax.tree as jt
from jaxtyping import Array, ArrayLike, Float, Integer, PRNGKeyArray, Real, Scalar

from bayinx.core.distribution import Distribution
from bayinx.core.node import Node
from bayinx.nodes import Observed


def log_binom_coeff(n: ArrayLike, x: ArrayLike) -> Array:
    n, x = jnp.asarray(n), jnp.asarray(x)
    return jsp.gammaln(n + 1) - jsp.gammaln(x + 1) - jsp.gammaln(n - x + 1)


def prob(
    x: Integer[ArrayLike, "..."],
    n: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The probability mass function (PMF) for a Binomial distribution.

    # Parameters
    - `x`: Where to evaluate the PMF (number of successes).
    - `n`: The number of trials.
    - `p`: The probability of success.

    # Returns
    The PMF evaluated at `x`. The output will have the broadcasted shapes of `x`, `n`, and `p`.
    """
    # Cast to Array
    x, n, p = jnp.asarray(x), jnp.asarray(n), jnp.asarray(p)

    return lax.exp(logprob(x, n, p))


def logprob(
    x: Integer[ArrayLike, "..."],
    n: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the probability mass function (log PMF) for a Binomial distribution.

    # Parameters
    - `x`: Where to evaluate the log PMF (number of successes).
    - `n`: The number of trials.
    - `p`: The probability of success.

    # Returns
    The log PMF evaluated at `x`. The output will have the broadcasted shapes of `x`, `n`, and `p`.
    """
    # Cast to Array
    k, n, p = jnp.asarray(x), jnp.asarray(n), jnp.asarray(p)

    return log_binom_coeff(n, k) + k * lax.log(p) + (n - k) * lax.log1p(-p)


def cdf(
    x: Integer[ArrayLike, "..."],
    n: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The cumulative density function (CDF) for a Binomial distribution.

    # Parameters
    - `x`: Where to evaluate the CDF (P(X <= x)).
    - `n`: The number of trials.
    - `p`: The probability of success.

    # Returns
    The CDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `n`, and `p`.
    """
    # Cast to Array
    x, n, p = jnp.asarray(x), jnp.asarray(n), jnp.asarray(p)

    return jsp.betainc(n - x, x + 1, 1.0 - p)


def logcdf(
    x: Integer[ArrayLike, "..."],
    n: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the cumulative density function (log CDF) for a Binomial distribution.

    # Parameters
    - `x`: Where to evaluate the log CDF (ln P(X <= x)).
    - `n`: The number of trials.
    - `p`: The probability of success.

    # Returns
    The log CDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `n`, and `p`.
    """
    # Cast to Array
    x, n, p = jnp.asarray(x), jnp.asarray(n), jnp.asarray(p)

    return lax.log(cdf(x, n, p))


def ccdf(
    x: Integer[ArrayLike, "..."],
    n: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The complementary cumulative density function (cCDF) for a Binomial distribution (P(X > k)).

    # Parameters
    - `k`: Where to evaluate the cCDF.
    - `n`: The number of trials.
    - `p`: The probability of success.

    # Returns
    The cCDF evaluated at `k`. The output will have the broadcasted shapes of `k`, `n`, and `p`.
    """

    # Cast to Array
    x, n, p = jnp.asarray(x), jnp.asarray(n), jnp.asarray(p)

    return jsp.betainc(x + 1, n - x, p)


def logccdf(
    x: Integer[ArrayLike, "..."],
    n: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the complementary cumulative density function (log cCDF) for a Binomial distribution.

    # Parameters
    - `x`: Where to evaluate the log cCDF.
    - `n`: The number of trials.
    - `p`: The probability of success.

    # Returns
    The log cCDF evaluated at `x`. The output will have the broadcasted shapes of `x`, `n`, and `p`.
    """
    # Cast to Array
    x, n, p = jnp.asarray(x), jnp.asarray(n), jnp.asarray(p)

    return lax.log(ccdf(x, n, p))


class Binomial(Distribution):
    """
    A Binomial distribution.

    # Attributes
    - `n`: The number of trials parameter (integer).
    - `p`: The probability of success parameter (float, [0, 1]).
    """

    n: Node[Integer[Array, "..."]]
    p: Node[Real[Array, "..."]]


    def __init__(
        self,
        n: Integer[ArrayLike, "..."] | Node[Integer[Array, "..."]],
        p: Real[ArrayLike, "..."] | Node[Real[Array, "..."]]
    ):
        # Initialize number of trials (n)
        if isinstance(n, Node):
            if isinstance(n.obj, ArrayLike):
                self.n = n # type: ignore
        else:
            self.n = Observed(jnp.asarray(n))

        # Initialize probability of success parameter (p)
        if isinstance(p, Node):
            if isinstance(p.obj, ArrayLike):
                self.p = p # type: ignore
        else:
            self.p = Observed(jnp.asarray(p))

    def logprob(self, node: Node) -> Scalar:
        obj, n, p = (
            node.obj,
            self.n.obj,
            self.p.obj
        )

        # Filter out irrelevant values
        obj, _ = eqx.partition(obj, node._filter_spec)

        # Helper function for the single-leaf log-probability evaluation
        def leaf_logprob(k: Float[ArrayLike, "..."]) -> Scalar:
            return logprob(k, n, p).sum()

        # Compute log probabilities across leaves
        eval_obj = jt.map(leaf_logprob, obj)

        # Compute total sum
        total = jt.reduce_associative(lambda x,y: x + y, eval_obj, identity=0.0)

        return jnp.asarray(total)

    def sample(self, shape: int | Tuple[int, ...], key: PRNGKeyArray = jr.key(0)):
        # Coerce to tuple
        if isinstance(shape, int):
            shape = (shape, )

        # Use jr.binomial for sampling (returns integer array)
        # Note: jr.binomial accepts n as int, p as float, and shape
        return jr.binomial(key, self.n.obj, self.p.obj, shape=shape)



from typing import Protocol, Tuple

import jax.random as jr
from jaxtyping import PRNGKeyArray, Scalar

from bayinx.core.node import Node


class Distribution(Protocol):
    """
    A protocol used for defining the structure of distributions.
    """

    def logprob(self, node: Node) -> Scalar: ...

    def sample(self, shape: int | Tuple[int, ...], key: PRNGKeyArray = jr.key(0)): ...

    def __rlshift__(self, node: Node):
        """
        Implicitly accumulate the log probability into the current model context.
        """
        from bayinx.core.context import _model_context

        # Evaluate log posterior
        log_prob = self.logprob(node)

        # Accumulate log probability into context
        if hasattr(_model_context, "target"):
            _model_context.target += log_prob
        else:
            raise RuntimeError(
                "Model context doesn't exist. Make sure you're calling "
                + "this within the 'model' method."
            )



from typing import Tuple

import equinox as eqx
import jax.lax as lax
import jax.numpy as jnp
import jax.random as jr
import jax.tree as jt
from jaxtyping import Array, ArrayLike, PRNGKeyArray, Real, Scalar

from bayinx.core.distribution import Distribution
from bayinx.core.node import Node
from bayinx.nodes import Observed


def prob(
    x: Real[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The probability density function (PDF) for an Exponential distribution.

    # Parameters
    - `x`: Where to evaluate the PDF (must be >= 0).
    - `lam`: The rate parameter (lambda), must be > 0.

    # Returns
    The PDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = lam * lax.exp(-lam * x)

    # Handle values outside of support
    result = lax.select(x >= 0, result, jnp.array(0.0))

    return result


def logprob(
    x: Real[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the probability density function (log PDF) for an Exponential distribution.

    # Parameters
    - `x`: Where to evaluate the log PDF (must be >= 0).
    - `lam`: The rate parameter (lambda), must be > 0.

    # Returns
    The log PDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = lax.log(lam) - lam * x

    # Handle values outside of support
    result = lax.select(x >= 0, result, -jnp.inf)

    return result

def cdf(
    x: Real[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The cumulative density function (CDF) for an Exponential distribution (P(X <= x)).

    # Parameters
    - `x`: Where to evaluate the CDF.
    - `lam`: The rate parameter (lambda).

    # Returns
    The CDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = 1.0 - lax.exp(-lam * x)

    # Handle values outside of support
    result = lax.select(x >= 0, result, jnp.array(0.0))

    return result

def logcdf(
    x: Real[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the cumulative density function (log CDF) for an Exponential distribution.

    # Parameters
    - `x`: Where to evaluate the log CDF (ln P(X <= x)).
    - `lam`: The rate parameter (lambda).

    # Returns
    The log CDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = lax.log1p(-lax.exp(-lam * x))

    # Handle values outside of support (x < 0)
    result = lax.select(x >= 0, result, -jnp.inf)

    return result


def ccdf(
    x: Real[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The complementary cumulative density function (cCDF) for an Exponential distribution (P(X > x)).

    # Parameters
    - `x`: Where to evaluate the cCDF.
    - `lam`: The rate parameter (lambda).

    # Returns
    The cCDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = lax.exp(-lam * x)

    # Handle values outside of support
    result = lax.select(x >= 0, result, jnp.array(1.0))

    return result


def logccdf(
    x: Real[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the complementary cumulative density function (log cCDF) for an Exponential distribution.

    # Parameters
    - `x`: Where to evaluate the log cCDF.
    - `lam`: The rate parameter (lambda).

    # Returns
    The log cCDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    # log(cCDF(x)) = -lambda * x for x >= 0
    log_ccdf_val = -lam * x

    # Handle values outside of support (x < 0), log(P(X > x)) = log(1.0) = 0.0
    return lax.select(x >= 0, log_ccdf_val, jnp.array(0.0))


class Exponential(Distribution):
    """
    An Exponential distribution.

    # Attributes
    - `lam`: The rate parameter (lambda), must be positive.
    """

    lam: Node[Real[Array, "..."]]


    def __init__(
        self,
        lam: Real[ArrayLike, "..."] | Node[Real[Array, "..."]],
    ):
        # Initialize rate parameter (lambda)
        if isinstance(lam, Node):
            if isinstance(lam.obj, ArrayLike):
                self.lam = lam # type: ignore
        else:
            self.lam = Observed(jnp.asarray(lam))


    def logprob(self, node: Node) -> Scalar:
        obj, lam = (
            node.obj,
            self.lam.obj,
        )

        # Filter out irrelevant values
        obj, _ = eqx.partition(obj, node._filter_spec)

        # Helper function for the single-leaf log-probability evaluation
        def leaf_logprob(x: Real[ArrayLike, "..."]) -> Scalar:
            return logprob(x, lam).sum()

        # Compute log probabilities across leaves
        eval_obj = jt.map(leaf_logprob, obj)

        # Compute total sum
        total = jt.reduce_associative(lambda x,y: x + y, eval_obj, identity=0.0)

        return jnp.asarray(total)

    def sample(self, shape: int | Tuple[int, ...], key: PRNGKeyArray = jr.key(0)):
        # Coerce to tuple
        if isinstance(shape, int):
            shape = (shape, )

        return jr.exponential(key, shape=shape) *  1.0 / self.lam.obj



from .bernoulli import Bernoulli as Bernoulli
from .binomial import Binomial as Binomial
from .exponential import Exponential as Exponential
from .normal import Normal as Normal
from .poisson import Poisson as Poisson



import jax.numpy as jnp
from jaxtyping import Array, Scalar

from bayinx import Model, Posterior, define
from bayinx.dists import Normal
from bayinx.flows import DiagAffine, FullAffine
from bayinx.nodes import Continuous, Observed


class ShiftedNormal(Model, init=False):
    mu: Continuous[Scalar] = define(shape = ())
    y: Observed[Array] = define()

    def model(self, target):
        self.y << Normal(self.mu, 1.0)

        return target

class CenteredNormal(Model):
    sigma: Continuous[Scalar] = define(shape = (), lower = 0.0)
    y: Observed[Array] = define()

    def model(self, target):
        self.y << Normal(0.0, self.sigma)

        return target

class ArbitraryNormal(Model):
    mu: Continuous[Scalar] = define(shape = ())
    sigma: Continuous[Scalar] = define(shape = (), lower = 0.0)
    y: Observed[Array] = define()

    def model(self, target):
        self.y << Normal(self.mu, self.sigma)

        return target

y_data: Array = jnp.array([-1.0, 0.0, 1.0])

def test_shifted():
    # Construct and configure posterior
    posterior = Posterior(ShiftedNormal, y = y_data)
    posterior.configure(flowspecs = [DiagAffine()])

    # Fit variational approximation
    posterior.fit(max_iters = 1000)
    posterior.fit(max_iters = 1000, grad_draws = 4)
    posterior.fit(max_iters = 10000, grad_draws = 6, batch_size = 3, learning_rate=1e-3)

    # Generate posterior samples
    mu_post = posterior.sample(node = 'mu', n_draws = 10000)

    assert mu_post.shape == (10000, )



def test_scaled():
    y_data: Array = jnp.array([-1.0, 0.0, 1.0])

    posterior = Posterior(CenteredNormal, y = y_data)
    posterior.configure(flowspecs = [DiagAffine()])
    posterior.fit(max_iters = 1000)
    posterior.fit(max_iters = 1000, grad_draws = 4)
    posterior.fit(max_iters = 1000, grad_draws = 6, batch_size = 3)

    # Generate posterior samples
    sigma_post = posterior.sample(node = 'sigma', n_draws = 10000)

    assert sigma_post.shape == (10000, )

def test_arbitrary():
    y_data: Array = jnp.array([-1.0, 0.0, 1.0])

    posterior = Posterior(ArbitraryNormal, y = y_data)
    posterior.configure(flowspecs = [FullAffine()])
    posterior.fit(max_iters = 1000)
    posterior.fit(max_iters = 1000, grad_draws = 4)
    posterior.fit(max_iters = 1000, grad_draws = 6, batch_size = 3)

    # Generate posterior samples
    mu_post = posterior.sample(node = 'mu', n_draws = 10000)
    sigma_post = posterior.sample(node = 'sigma', n_draws = 10000)

    assert mu_post.shape == (10000, )
    assert sigma_post.shape == (10000, )



from typing import Callable, Dict, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Array, Float, PRNGKeyArray, Scalar

from bayinx.core.flow import FlowLayer, FlowSpec


class FullAffineLayer(FlowLayer):
    """
    A full affine flow.

    # Attributes
    - `params`: The parameters of the full affine flow.
    - `constraints`: The constraining transformations for the parameters of the full affine flow.
    - `static`: Whether the flow layer is frozen (parameters are not subject to further optimization).
    """

    params: Dict[str, Array]
    constraints: Dict[str, Callable[[Array], Array]]
    static: bool


    def __init__(self, dim: int, key: PRNGKeyArray):
        """
        Initializes a full affine flow.

        # Parameters
        - `dim`: The dimension of the parameter space.
        """
        self.static = False

        # Split key
        k1, k2 = jr.split(key)

        # Initialize parameters
        self.params = {
            "shift": jr.normal(key, (dim, )) / dim**0.5,
            "scale": jr.normal(key, (dim, dim)) / dim**0.5,
        }

        # Define constraints
        if dim == 1:
            self.constraints = {"scale": jnp.exp}
        else:
            def constrain_scale(scale: Array):
                # Extract diagonal and apply exponential
                diag: Array = jnp.exp(jnp.diag(scale))

                # Return matrix with modified diagonal
                return jnp.fill_diagonal(jnp.tril(scale), diag, inplace=False)

            self.constraints = {"scale": constrain_scale}

    @eqx.filter_jit
    def forward(self, draws: Float[Array, "n_draws n_dim"]) -> Float[Array, "n_draws n_dim"]:
        params = self.transform_params()

        # Extract parameters
        shift: Float[Array, " n_dim"] = params["shift"]
        scale: Float[Array, "n_dim n_dim"] = params["scale"]

        # Compute forward transformation
        draws = (scale @ draws.T).T + shift

        return draws

    def __adjust(self, draw: Float[Array, " n_dim"]) -> Float[Array, " n_dim"]:
        params = self.transform_params()

        # Extract parameters
        scale: Float[Array, "n_dim n_dim"] = params["scale"]

        # Compute log-Jacobian adjustments
        lja: Array = jnp.log(jnp.diag(scale)).sum()

        assert lja.shape == ()

        return lja

    @eqx.filter_jit
    def adjust(self, draws: Float[Array, "n_draws n_dim"]) -> Float[Array, "n_draws n_dim"]:
        f = jax.vmap(self.__adjust, 0)
        return f(draws)

    def __forward_and_adjust(self, draw: Float[Array, " n_dim"]) -> Tuple[Float[Array, " n_dim"], Scalar]:
        params = self.transform_params()

        assert len(draw.shape) == 1

        # Extract parameters
        shift: Float[Array, " n_dim"] = params["shift"]
        scale: Float[Array, "n_dim n_dim"] = params["scale"]

        # Compute forward transformation
        draw = (scale @ draw.T).T + shift

        assert len(draw.shape) == 1

        # Compute lja
        lja: Scalar = jnp.log(jnp.diag(scale)).sum()

        assert lja.shape == ()

        return draw, lja

    @eqx.filter_jit
    def forward_and_adjust(self, draws: Float[Array, "n_draws n_dim"]) -> Tuple[Float[Array, "n_draws n_dim"], Scalar]:
        f = jax.vmap(self.__forward_and_adjust, 0)
        return f(draws)


class FullAffine(FlowSpec):
    key: PRNGKeyArray

    def __init__(self, key: PRNGKeyArray = jr.key(0)):
        self.key = key

    def construct(self, dim: int) -> FullAffineLayer:
        return FullAffineLayer(dim, self.key)



from bayinx.flows.diagaffine import DiagAffine as DiagAffine
from bayinx.flows.fullaffine import FullAffine as FullAffine
from bayinx.flows.lowrankaffine import LowRankAffine as LowRankAffine
from bayinx.flows.planar import Planar as Planar
from bayinx.flows.sylvester import Sylvester as Sylvester



"""
Test inference for a linear model:

    Y_i ~ Normal(x_i * beta, sigma)
"""

import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Array, Scalar

import bayinx as byx
from bayinx import define
from bayinx.dists import Normal
from bayinx.flows import LowRankAffine
from bayinx.nodes import Continuous, Observed


# Define model
class LinearModel(byx.Model):
    beta: Continuous[Scalar] = define(shape = 'n_predictors')
    sigma: Continuous[Scalar] = define(shape = (), lower = 0.0)

    X: Observed[Array] = define(shape = ('n_obs', 'n_predictors'))
    y: Observed[Array] = define(shape = 'n_obs')

    def model(self, target):
        # Compute expected response
        mu = self.X @ self.beta

        # Accumulate likelihood
        self.y << Normal(mu, self.sigma)

        return target

# Simulate sample
n_obs = 2500
n_predictors = 5
X: Array = jr.normal(jr.key(0), (n_obs, n_predictors - 1))
X = jnp.column_stack((jnp.ones((n_obs,)), X))
beta = jnp.array(range(n_predictors))

y = jr.normal(jr.key(0), (n_obs, )) * 0.5 + X @ beta

def test_inference():
    # Define posterior
    posterior = byx.Posterior(LinearModel,
        n_obs = n_obs,
        n_predictors = n_predictors,
        X = X,
        y = y
    )

    # Configure and fit
    posterior.configure(flowspecs = [LowRankAffine(2)])
    posterior.fit(max_iters = int(1e5), learning_rate = 1e-2)

    # Check fit
    assert jnp.linalg.norm(posterior.sample('beta', int(1e6)).mean(0) - beta) < 0.1



from typing import Any, Generic, Iterator, Self

import equinox as eqx
import jax.tree as jt
from jaxtyping import PyTree

from bayinx.core.types import T
from bayinx.core.utils import _extract_obj, _merge_filter_specs


class Node(eqx.Module, Generic[T]):
    """
    A thin wrapper for nodes of a probabilistic model.


    # Attributes
    - `obj`: An internal realization of the node.
    - `_filter_spec`: An internal filter specification for `obj`.
    """

    obj: T
    _filter_spec: PyTree

    def __init__(self, obj, filter_spec):
        self.obj = obj
        self._filter_spec = filter_spec

    @property
    def filter_spec(self) -> Self:
        """
        An outer filter specification for the full node.
        """
        # Generate empty specification
        node_filter_spec: Self = jt.map(lambda _: False, self)

        # Filter based on inner filter specification for 'obj'
        node_filter_spec = eqx.tree_at(
            lambda node: node.obj,
            node_filter_spec,
            replace=self._filter_spec,
        )

        return node_filter_spec

    # Wrappers around internal dunder methods ----
    def __getitem__(self, key: Any) -> "Node":
        if isinstance(key, Node):
            raise TypeError("Subsetting nodes with nodes is not yet supported.")

        # Subset internally
        new_obj = self.obj[key]
        if type(self.obj) is type(self._filter_spec):
            new_filter_spec = self._filter_spec[key]
        else:
            new_filter_spec = self._filter_spec

        # Create new subsetted node
        return type(self)(new_obj, new_filter_spec)

    def __iter__(self) -> Iterator["Node"]:
        for obj_i, spec_i in zip(self.obj, self._filter_spec):
            # Create a new Node for the current element
            yield Node(obj_i, spec_i)


    ## Arithmetic ----
    def __add__(self, other: Any) -> "Node":
        # Extract internal objects and their filter specifications
        lhs_obj, lhs_filter_spec = _extract_obj(self)
        rhs_obj, rhs_filter_spec = _extract_obj(other)

        # Perform addition
        new_obj = lhs_obj + rhs_obj

        # Merge filter specifications
        new_filter_spec = _merge_filter_specs(
            [lhs_filter_spec, rhs_filter_spec],
            [lhs_obj, rhs_obj],
            new_obj
        )

        return Node(new_obj, new_filter_spec)

    def __sub__(self, other: Any) -> "Node":
        # Extract internal objects and their filter specifications
        lhs_obj, lhs_filter_spec = _extract_obj(self)
        rhs_obj, rhs_filter_spec = _extract_obj(other)

        # Perform subtraction
        new_obj = lhs_obj - rhs_obj

        # Merge filter specifications
        new_filter_spec = _merge_filter_specs(
            [lhs_filter_spec, rhs_filter_spec],
            [lhs_obj, rhs_obj],
            new_obj
        )

        return Node(new_obj, new_filter_spec)

    def __mul__(self, other: Any) -> "Node":
        # Extract internal objects and their filter specifications
        lhs_obj, lhs_filter_spec = _extract_obj(self)
        rhs_obj, rhs_filter_spec = _extract_obj(other)

        # Perform multiplication
        new_obj = lhs_obj * rhs_obj

        # Merge filter specifications
        new_filter_spec = _merge_filter_specs(
            [lhs_filter_spec, rhs_filter_spec],
            [lhs_obj, rhs_obj],
            new_obj
        )

        return Node(new_obj, new_filter_spec)

    def __matmul__(self, other: Any) -> "Node":
        # Extract internal objects and their filter specifications
        lhs_obj, lhs_filter_spec = _extract_obj(self)
        rhs_obj, rhs_filter_spec = _extract_obj(other)

        # Perform matrix multiplication
        new_obj = lhs_obj @ rhs_obj

        # Merge filter specifications
        new_filter_spec = _merge_filter_specs(
            [lhs_filter_spec, rhs_filter_spec],
            [lhs_obj, rhs_obj],
            new_obj
        )

        return Node(new_obj, new_filter_spec)

    def __truediv__(self, other: Any) -> "Node":
        # Extract internal objects and their filter specifications
        lhs_obj, lhs_filter_spec = _extract_obj(self)
        rhs_obj, rhs_filter_spec = _extract_obj(other)

        # Perform true division
        new_obj = lhs_obj / rhs_obj

        # Merge filter specifications
        new_filter_spec = _merge_filter_specs(
            [lhs_filter_spec, rhs_filter_spec],
            [lhs_obj, rhs_obj],
            new_obj
        )

        return Node(new_obj, new_filter_spec)

    def __floordiv__(self, other: Any) -> "Node":
        # Extract internal objects and their filter specifications
        lhs_obj, lhs_filter_spec = _extract_obj(self)
        rhs_obj, rhs_filter_spec = _extract_obj(other)

        # Perform floor division
        new_obj = lhs_obj // rhs_obj

        # Merge filter specifications
        new_filter_spec = _merge_filter_specs(
            [lhs_filter_spec, rhs_filter_spec],
            [lhs_obj, rhs_obj],
            new_obj
        )

        return Node(new_obj, new_filter_spec)

    def __pow__(self, other: Any) -> "Node":
         # Extract internal objects and their filter specifications
        lhs_obj, lhs_filter_spec = _extract_obj(self)
        rhs_obj, rhs_filter_spec = _extract_obj(other)

        # Perform floor division
        new_obj = lhs_obj ** rhs_obj

        # Merge filter specifications
        new_filter_spec = _merge_filter_specs(
            [lhs_filter_spec, rhs_filter_spec],
            [lhs_obj, rhs_obj],
            new_obj
        )

        return Node(new_obj, new_filter_spec)

    def __mod__(self, other: Any) -> "Node":
        # Extract internal objects and their filter specifications
        lhs_obj, lhs_filter_spec = _extract_obj(self)
        rhs_obj, rhs_filter_spec = _extract_obj(other)

        # Perform modulus
        new_obj = lhs_obj % rhs_obj

        # Merge filter specifications
        new_filter_spec = _merge_filter_specs(
            [lhs_filter_spec, rhs_filter_spec],
            [lhs_obj, rhs_obj],
            new_obj
        )

        return Node(new_obj, new_filter_spec)



from typing import Tuple

import equinox as eqx
import jax.lax as lax
import jax.numpy as jnp
import jax.random as jr
import jax.tree as jt
from jaxtyping import Array, ArrayLike, Float, Integer, PRNGKeyArray, Real, Scalar

from bayinx.core.distribution import Distribution
from bayinx.core.node import Node
from bayinx.nodes import Observed


def prob(
    x: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."]
) -> Real[Array, "..."]:
    """
    The probability mass function (PMF) for a Bernoulli distribution.

    # Parameters
    - `x`: Where to evaluate the PMF (pst be 0 or 1).
    - `p`: The probability of success (p).

    # Returns
    The PMF evaluated at `x`. The output will have the broadcasted shapes of `x` and `p`.
    """
    # Cast to Array
    x, p = jnp.asarray(x), jnp.asarray(p)

    # Bernoulli PMF: p^x * (1-p)^(1-x)
    return lax.exp(x * lax.log(p) + (1.0 - x) * lax.log1p(-p))


def logprob(
    x: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."]
) -> Real[Array, "..."]:
    """
    The log of the probability mass function (log PMF) for a Bernoulli distribution.

    # Parameters
    - `x`: Where to evaluate the log PMF (pst be 0 or 1).
    - `p`: The probability of success (p).

    # Returns
    The log PMF evaluated at `x`. The output will have the broadcasted shapes of `x` and `p`.
    """
    # Cast to Array
    x, p = jnp.asarray(x), jnp.asarray(p)

    return x * lax.log(p) + (1.0 - x) * lax.log(1.0 - p)


def cdf(
    x: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."]
) -> Real[Array, "..."]:
    """
    The cuplative distribution function (CDF) for a Bernoulli distribution.

    # Parameters
    - `x`: Where to evaluate the CDF.
    - `p`: The probability of success (p).

    # Returns
    The CDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `p`.
    """
    # Cast to Array
    x, p = jnp.asarray(x), jnp.asarray(p)

    return jnp.where(
        x < 0.0,
        0.0,
        jnp.where(x < 1.0, 1.0 - p, 1.0)
    )


def logcdf(
    x: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."]
) -> Real[Array, "..."]:
    """
    The log of the cuplative distribution function (log CDF) for a Bernoulli distribution.

    # Parameters
    - `x`: Where to evaluate the log CDF.
    - `p`: The probability of success (p).

    # Returns
    The log CDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `p`.
    """
    # Cast to Array
    x, p = jnp.asarray(x), jnp.asarray(p)

    return jnp.where(
        x < 0.0,
        -jnp.inf,
        jnp.where(x < 1.0, lax.log1p(-p), 0.0)
    )


def ccdf(
    x: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."]
) -> Real[Array, "..."]:
    """
    The complementary cuplative distribution function (cCDF) for a Bernoulli distribution.

    # Parameters
    - `x`: Where to evaluate the cCDF.
    - `p`: The probability of success (p).

    # Returns
    The cCDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `p`.
    """

    # Cast to Array
    x, p = jnp.asarray(x), jnp.asarray(p)

    return jnp.where(
        x < 0.0,
        1.0,
        jnp.where(x < 1.0, p, 0.0)
    )


def logccdf(
    x: Integer[ArrayLike, "..."],
    p: Real[ArrayLike, "..."]
) -> Real[Array, "..."]:
    """
    The log of the complementary cuplative distribution function (log cCDF) for a Bernoulli distribution.

    # Parameters
    - `x`: Where to evaluate the log cCDF.
    - `p`: The probability of success (p).

    # Returns
    The log cCDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `p`.
    """
    # Cast to Array
    x, p = jnp.asarray(x), jnp.asarray(p)

    return jnp.where(
        x < 0.0,
        0.0,
        jnp.where(x < 1.0, lax.log(p), -jnp.inf)
    )


class Bernoulli(Distribution):
    """
    A Bernoulli distribution.

    # Attributes
    - `p`: The probability of success (p).
    """

    p: Node[Real[Array, "..."]]


    def __init__(
        self,
        p: Real[ArrayLike, "..."] | Node[Real[ArrayLike, "..."]],
    ):
        # Initialize probability of success parameter (p)
        if isinstance(p, Node):
            if isinstance(p.obj, ArrayLike):
                self.p = p # type: ignore
        else:
            self.p = Observed(jnp.asarray(p))


    def logprob(self, node: Node) -> Scalar:
        obj, p = (
            node.obj,
            self.p.obj,
        )

        # Filter out irrelevant values
        obj, _ = eqx.partition(obj, node._filter_spec)

        # Helper function for the single-leaf log-probability evaluation
        def leaf_logprob(x: Float[ArrayLike, ""]) -> Scalar:
            return logprob(x, p).sum()

        # Compute log probabilities across leaves
        eval_obj = jt.map(leaf_logprob, obj)

        # Compute total sum
        total = jt.reduce_associative(lambda x,y: x + y, eval_obj, identity=0.0)

        return jnp.asarray(total)

    def sample(self, shape: int | Tuple[int, ...], key: PRNGKeyArray = jr.key(0)):
        # Coerce to tuple
        if isinstance(shape, int):
            shape = (shape, )

        return jr.bernoulli(key, self.p.obj, shape)




from typing import Any, Dict, List, Optional, Tuple

import jax.tree as jt
from jaxtyping import PyTree


def _extract_shape_params(shape_spec: int | str | Tuple[int | str, ...]) -> set[str]:
    """
    Extract parameter names from shape specification.
    """
    params = set()
    if isinstance(shape_spec, str):
        params.add(shape_spec)
    elif isinstance(shape_spec, tuple):
        for item in shape_spec:
            if isinstance(item, str):
                params.add(item)
    #
    return params

def _resolve_shape_spec(
    shape_spec: None | int | str | Tuple[int | str, ...],
    shape_values: Dict[str, int]
) -> None | Tuple[int, ...]:
    """
    Replaces named dimensions in a shape specification with their integer or tuple values.

    # Example
    For `shape_values = {'k': 5, 's': (3,2,1)}`:
        `('k', 4, 's')` --> `(5, 4, 3, 2, 1)`
    """
    if shape_spec is None:
        return None

    # Coerce to tuple for uniform processing
    if isinstance(shape_spec, (int, str)):
        shape_spec = (shape_spec,)

    resolved_spec: List[int] = []
    for dim in shape_spec:
        if isinstance(dim, str):
            if dim in shape_values:
                resolved_value = shape_values[dim] # Grab initialized value

                if isinstance(resolved_value, int):
                    # Scalar shape dimension (e.g., 'k' -> 3)
                    resolved_spec.append(resolved_value)
                elif isinstance(resolved_value, tuple):
                    # Packed shape dimension (e.g., 'shape' -> (3, 2, 1))
                    resolved_spec.extend(resolved_value)
                else:
                    raise TypeError(f"Shape parameter '{dim}' resolved to an unsupported type: {type(resolved_value).__name__}")
            else: # dim not in shape_values
                raise TypeError(f"Shape parameter '{dim}' was not initialized with a value.")
        elif isinstance(dim, int):
            # Literal integer (e.g., 3 -> 3)
            resolved_spec.append(dim)
        else:
            raise TypeError(f"Shape parameter {dim} was incorrectly specified (must be 'int' or 'str', got '{type(dim).__name__}').")

    return tuple(resolved_spec)

def _extract_obj(x: Any) -> Tuple[Any, Any]:
    """
    Extract the object and its (potentially implicit) filter specification.
    """
    from bayinx.core.node import Node

    if isinstance(x, Node):
        obj = x.obj
        filter_spec = x._filter_spec
    else:
        obj: Any = x # type: ignore
        filter_spec = True # implicit filter specification

    return (obj, filter_spec)


def _merge_filter_specs(
    filter_specs: List[PyTree],
    objs: Optional[List[PyTree]] = None,
    obj: Optional[PyTree] = None
) -> PyTree:
    """
    Merge filter specifications that share the type of `obj`.

    If `obj` and `objs` are not provided then `filter_specs` will be merged as is.
    """

    def _merge(*args):
        return all(args)

    if objs is None and obj is None:
        filter_spec: Any = jt.map(_merge, *filter_specs) # type: ignore
    else:
        selected_specs: List[PyTree] = []

        # Include filter specs whose objects share the correct type
        for cur_obj, cur_spec in zip(objs, filter_specs): # type: ignore
            if type(obj) is type(cur_obj):
                selected_specs.append(cur_spec)

        if len(selected_specs) != 0:
            filter_spec = jt.map(_merge, *selected_specs)
        else:
            filter_spec = True

    return filter_spec



from typing import Any, Tuple

import jax.numpy as jnp
import jax.tree as jt
from jaxtyping import Scalar, ScalarLike

from bayinx.core.constraint import Constraint
from bayinx.core.types import T


class Lower(Constraint):
    """
    Enforces a lower bound on the parameter.

    # Attributes
    - `lb`: The lower bound.
    """

    lb: Scalar

    def __init__(self, lb: ScalarLike):
        self.lb = jnp.asarray(lb)

    def constrain(self, obj: T, filter_spec: T) -> Tuple[T, Scalar]:
        """
        Applies the exponential transformation to the leaves of a `PyTree` and
        computes the log-Jacobian adjustment of the transformation.

        # Parameters
        - `x`: The unconstrained `PyTree`.

        # Returns
        A tuple containing:
            - A `PyTree` with its values `x` now satisfying `lb <= x`.
            - A scalar `Array` containing the log-absolute-Jacobian of the
                transformation.
        """
        log_jac: Scalar = jnp.array(0.0)

        def constrain_leaf(leaf: Any, filter: bool):
            nonlocal log_jac  # Reference outer variable

            if filter:
                # Apply transformation
                constrained = jnp.exp(leaf) + self.lb

                # Accumulate Jacobian adjustment
                log_jac = log_jac + jnp.sum(leaf)

                return constrained
            else:
                return leaf

        # Constrain leaves
        obj = jt.map(constrain_leaf, obj, filter_spec)

        return obj, log_jac



from typing import Callable, Dict, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Array, Float, PRNGKeyArray, Scalar

from bayinx.core.flow import FlowLayer, FlowSpec


class DiagAffineLayer(FlowLayer):
    """
    A diagonal (element-wise) affine flow.

    # Attributes
    - `params`: The parameters of the diagonal affine flow.
    - `constraints`: The constraining transformations for the parameters of the diagonal affine flow.
    - `static`: Whether the flow layer is frozen (parameters are not subject to further optimization).
    """

    params: Dict[str, Array]
    constraints: Dict[str, Callable[[Array], Array]]
    static: bool

    def __init__(self, dim: int, key: PRNGKeyArray):
        """
        Initializes a full affine flow.

        # Parameters
        - `dim`: The dimension of the parameter space.
        """
        self.static = False
        # Split key
        k1, k2 = jr.split(key)

        # Initialize parameters
        self.params = {
            "shift": jr.normal(k1, (dim,)) / dim**0.5,
            "scale": jr.normal(k2, (dim,)) / dim**0.5,
        }

        # Define constraints
        self.constraints = {"scale": jnp.exp}


    def __forward(self, draw: Float[Array, " D"]) -> Float[Array, " D"]:
        params = self.transform_params()

        assert len(draw.shape) == 1

        # Extract parameters
        shift: Float[Array, " dim"] = params["shift"]
        scale: Float[Array, " dim"] = params["scale"]

        # Compute forward transformation
        draw = draw * scale + shift

        return draw

    @eqx.filter_jit
    def forward(self, draws: Float[Array, "draws dim"]) -> Float[Array, "draws dim"]:
        f = jax.vmap(self.__forward, 0)
        return f(draws)

    def __adjust(self, draw: Float[Array, " dim"]) -> Float[Array, " dim"]:
        params = self.transform_params()

        # Extract parameters
        scale: Float[Array, " dim"] = params["scale"]

        # Compute log-Jacobian adjustment
        lja: Array = jnp.log(scale).sum()

        assert lja.shape == ()

        return lja

    @eqx.filter_jit
    def adjust(self, draws: Float[Array, "draws dim"]) -> Float[Array, "draws dim"]:
        f = jax.vmap(self.__adjust, 0)
        return f(draws)

    def __forward_and_adjust(self, draw: Float[Array, " dim"]) -> Tuple[Float[Array, " dim"], Scalar]:
        params = self.transform_params()

        assert len(draw.shape) == 1

        # Extract parameters
        shift: Float[Array, " dim"] = params["shift"] # noqa
        scale: Float[Array, " dim"] = params["scale"] # noqa

        # Compute forward transformation
        draw = scale * draw + shift

        assert len(draw.shape) == 1

        # Compute log-Jacobian adjustment
        lja: Scalar = jnp.log(scale).sum()

        assert lja.shape == ()

        return draw, lja

    @eqx.filter_jit
    def forward_and_adjust(self, draws: Float[Array, "draws dim"]) -> Tuple[Float[Array, "draws dim"], Scalar]:
        f = jax.vmap(self.__forward_and_adjust, 0)
        return f(draws)


class DiagAffine(FlowSpec):
    """
    A specification for the diagonal affine flow.
    """
    key: PRNGKeyArray
    def __init__(self, key: PRNGKeyArray = jr.key(0)):
        self.key = key

    def construct(self, dim: int) -> DiagAffineLayer:
        return DiagAffineLayer(dim, self.key)



import os
import sys

# Insert the path to your source code directory (src/) into sys.path.
# This allows Sphinx to find and import your 'bayinx' package.
sys.path.insert(0, os.path.abspath('../../src'))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Bayinx'
copyright = '2025, Todd Pocuca'
author = 'Todd Pocuca'
release = '0.5.3'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# **EDITED**: Added essential extensions for API documentation and themes
extensions = [
    'sphinx.ext.autodoc',      # Imports and processes docstrings from Python modules
    'sphinx.ext.napoleon',     # Supports Google/NumPy style docstrings
    'sphinx.ext.viewcode',     # Adds links to the source code
]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# **EDITED**: Switched to the 'sphinx_rtd_theme'
html_theme = 'alabaster'
html_static_path = ['_static']



from abc import abstractmethod
from typing import Optional

from bayinx.core.node import Node
from bayinx.core.types import T


class Stochastic(Node[T]):
    """
    A container for stochastic (unobserved) nodes of a probabilistic model.

    Subclasses can be constructed with defined filter specifications (implement the `filter_spec` property).

    # Attributes
    - `obj`: An internal realization of the node.
    - `_filter_spec`: An internal filter specification for `obj`.
    """

    @abstractmethod
    def __init__(
        self,
        obj: T,
        filter_spec: Optional[T],
    ):
        pass



import jax.numpy as jnp
import jax.tree as jt
from jaxtyping import Array, ArrayLike, Real

from bayinx.core.node import Node
from bayinx.core.utils import _extract_obj


def exp(node: Node) -> Node:
    """
    Apply the exponential transformation (jnp.exp) to a node.
    """
    obj, filter_spec = _extract_obj(node)

    # Helper function for the single-leaf exponential transform
    def leaf_exp(x: Real[ArrayLike, "..."]) -> Array:
        return jnp.exp(x)

    # Apply exponential
    new_obj = jt.map(leaf_exp, obj)

    return type(node)(new_obj, filter_spec)


def log(node: Node) -> Node:
    """
    Apply the natural logarithm transformation (jnp.log) to a node.
    Handles input value restrictions (must be positive).
    """
    obj, filter_spec = _extract_obj(node)

    # Helper function for the single-leaf log transform
    def leaf_log(x: Real[ArrayLike, "..."]) -> Array:
        return jnp.log(x)

    # Apply logarithm
    new_obj = jt.map(leaf_log, obj)

    return type(node)(new_obj, filter_spec)


def sin(node: Node) -> Node:
    """
    Apply the sine transformation (jnp.sin) to a node.
    """
    obj, filter_spec = _extract_obj(node)

    # Helper function for the single-leaf sine transform
    def leaf_sin(x: Real[ArrayLike, "..."]) -> Array:
        return jnp.sin(x)

    # Apply sine
    new_obj = jt.map(leaf_sin, obj)

    return type(node)(new_obj, filter_spec)


def cos(node: Node) -> Node:
    """
    Apply the cosine transformation (jnp.cos) to a node.
    """
    obj, filter_spec = _extract_obj(node)

    # Helper function for the single-leaf cosine transform
    def leaf_cos(x: Real[ArrayLike, "..."]) -> Array:
        return jnp.cos(x)

    # Apply cosine
    new_obj = jt.map(leaf_cos, obj)

    return type(node)(new_obj, filter_spec)


def tanh(node: Node) -> Node:
    """
    Apply the hyperbolic tangent transformation (jnp.tanh) to a node.
    """
    obj, filter_spec = _extract_obj(node)

    # Helper function for the single-leaf tanh transform
    def leaf_tanh(x: Real[ArrayLike, "..."]) -> Array:
        return jnp.tanh(x)

    # Apply tanh
    new_obj = jt.map(leaf_tanh, obj)

    return type(node)(new_obj, filter_spec)


def sigmoid(node: Node) -> Node:
    """
    Apply the sigmoid (logistic) transformation to a node.
    Sigmoid formula: 1 / (1 + exp(-x))
    """
    obj, filter_spec = _extract_obj(node)

    # Helper function for the single-leaf sigmoid transform
    def leaf_sigmoid(x: Real[ArrayLike, "..."]) -> Array:
        return 1.0 / (1.0 + jnp.exp(-x)) # type: ignore

    # Apply sigmoid
    new_obj = jt.map(leaf_sigmoid, obj)

    return type(node)(new_obj, filter_spec)



from typing import Any, Tuple

import jax.numpy as jnp
import jax.tree as jt
from jaxtyping import Scalar, ScalarLike

from bayinx.core.constraint import Constraint
from bayinx.core.types import T


class Interval(Constraint):
    """
    Enforces that the parameter lies in the (lb, ub) interval using a scaled
    and shifted sigmoid transformation.

    # Attributes
    - `lb`: The lower bound.
    - `ub`: The upper bound.
    """

    lb: Scalar
    ub: Scalar

    def __init__(self, lb: ScalarLike, ub: ScalarLike):
        self.lb = jnp.asarray(lb)
        self.ub = jnp.asarray(ub)

    def constrain(self, obj: T, filter_spec: T) -> Tuple[T, Scalar]:
        """
        Applies the scaled Sigmoid transformation to the leaves of a `PyTree` and
        computes the log-Jacobian adjustment.

        # Parameters
        - `obj`: The unconstrained `PyTree` (values are in R).

        # Returns
        A tuple containing:
            - A `PyTree` with its values `y` now satisfying lb < y < ub.
            - A scalar `Array` containing the log-absolute-Jacobian of the
              transformation.
        """
        log_jac: Scalar = jnp.array(0.0)

        def constrain_leaf(leaf: Any, filter: bool):
            nonlocal log_jac  # Reference outer variable

            if filter:
                # Apply transformation
                constrained = self.lb + (self.ub - self.lb) * jnp.exp(leaf) / (1.0 + jnp.exp(leaf))

                log_jac = log_jac + (jnp.log(constrained - self.lb) +
                    jnp.log(self.ub - constrained) -
                    jnp.log(self.ub - self.lb)).sum()

                return constrained
            else:
                return leaf

        # Constrain leaves
        obj = jt.map(constrain_leaf, obj, filter_spec)

        return obj, log_jac



from abc import abstractmethod
from typing import TYPE_CHECKING, Tuple

import equinox as eqx
from jaxtyping import PyTree, Scalar

if TYPE_CHECKING:
    from bayinx.core.types import T


class Constraint(eqx.Module):
    """
    Abstract base class for defining constraints (for stochastic nodes).
    """

    @abstractmethod
    def constrain(self, obj: "T", filter_spec: PyTree) -> Tuple["T", Scalar]:
        """
        Applies the constraining transformation to the leaves of a `PyTree` and
        computes the log-Jacobian adjustment of the transformation.

        # Parameters
        - `x`: The unconstrained values.
        - `filter_spec`: The filter specification for `values`.

        # Returns
        A tuple containing:
            - A `PyTree` with its leaves now constrained.
            - A scalar `Array` containing the log-Jacobian adjustment of the
                transformation.
        """
        pass



from typing import Any, Optional

import equinox as eqx
import jax.numpy as jnp
import jax.tree as jt
import numpy as np
from jaxtyping import Array, PyTree

from bayinx.constraints import Identity
from bayinx.core.constraint import Constraint
from bayinx.core.types import T
from bayinx.nodes.stochastic import Stochastic


def is_float_like(element: Any) -> bool:
    """
    Check if `element` is float-like.

    The structure of this function is borrowed from the `Equinox` library.
    """
    if hasattr(element, "__jax_array__"):
        element = element.__jax_array__()
    if isinstance(element, (np.ndarray, np.generic)):
        return bool(np.issubdtype(element.dtype, np.floating))
    elif isinstance(element, Array):
        return jnp.issubdtype(element.dtype, jnp.floating)
    else:
        return isinstance(element, float)


class Continuous(Stochastic[T]):
    """
    A container for continuous stochastic nodes of a probabilistic model.


    # Attributes
    - `obj`: An internal realization of the node.
    - `_filter_spec`: An internal filter specification for `obj`.
    - `_constraint`: A constraining transformation.
    """

    _constraint: Constraint


    def __init__(
        self,
        obj: T,
        constraint: Constraint = Identity(),
        filter_spec: Optional[PyTree] = None
    ):
        if filter_spec is None: # Default filter specification
            # Generate empty specification
            filter_spec = jt.map(lambda _: False, obj)

            # Specify float-like leaves
            filter_spec = eqx.tree_at(
                where=lambda obj: obj,
                pytree=filter_spec,
                replace=jt.map(is_float_like, obj),
            )

        self.obj = obj
        self._filter_spec = filter_spec
        self._constraint = constraint



from abc import ABC, abstractmethod
from typing import Callable, Dict, Self, Tuple

import equinox as eqx
import jax.tree_util as jtu
from jaxtyping import Array, PyTree


class FlowLayer(eqx.Module):
    """
    An abstract base class for a flow layer.

    # Attributes
    - `params`: The parameters of the diffeomorphism. # TODO FOR ALL FLOWS
    - `constraints`: The constraining transformations for parameters.
    - `static`: Whether the flow layer is frozen (parameters are not subject to further optimization).
    """

    params: Dict[str, PyTree]
    constraints: Dict[str, Callable[[PyTree], Array]]
    static: bool

    @abstractmethod
    def forward(self, draws: Array) -> Array:
        """
        Computes the forward transformation of `draws`.
        """
        pass

    @abstractmethod
    def adjust(self, draws: Array) -> Array:
        """
        Computes the log-Jacobian adjustment for each draw in `draws`.

        # Returns
            An array of the log-Jacobian adjustments.
        """
        pass

    @abstractmethod
    def forward_and_adjust(self, draws: Array) -> Tuple[Array, Array]:
        """
        Computes the log-Jacobian adjustment at `draws` and applies the forward transformation.

        # Returns
            A tuple of JAX Arrays containing the transformed draws and log-absolute-Jacobians.
        """
        pass

    # Default filter specification
    @property
    def filter_spec(self) -> "FlowLayer":
        """
        Generates a filter specification to subset relevant parameters for the flow.
        """
        # Generate empty specification
        filter_spec = jtu.tree_map(lambda _: False, self)

        if self.static is False:
            # Specify parameters
            filter_spec = eqx.tree_at(
                lambda flow: flow.params,
                filter_spec,
                replace=True,
            )

        return filter_spec

    def constrain_params(self: Self) -> Dict[str, PyTree]:
        """
        Constrain flow parameters to the appropriate domain.

        # Returns
        The constrained parameters of the diffeomorphism.
        """
        params = self.params

        for par, map in self.constraints.items():
            params[par] = map(params[par])

        return params

    def transform_params(self: Self) -> Dict[str, PyTree]:
        """
        Apply a custom transformation to `params` if needed. Defaults to `constrain_params()`.

        # Returns
        The transformed parameters of the diffeomorphism.
        """
        return self.constrain_params()


class FlowSpec(ABC):
    """
    A specification for a flow layer.
    """

    @abstractmethod
    def construct(self, dim: int) -> FlowLayer: ...



from typing import Tuple

import equinox as eqx
import jax.lax as lax
import jax.numpy as jnp
import jax.random as jr
import jax.scipy.special as jsp
import jax.tree as jt
from jaxtyping import Array, ArrayLike, Integer, PRNGKeyArray, Real, Scalar

from bayinx.core.distribution import Distribution
from bayinx.core.node import Node
from bayinx.nodes import Observed


def prob(
    x: Integer[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The probability mass function (PMF) for a Poisson distribution.

    # Parameters
    - `x`: Where to evaluate the PMF.
    - `lam`: The rate parameter (lambda), representing the average number of events.

    # Returns
    The PMF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    return lax.exp(logprob(x, lam))


def logprob(
    x: Integer[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the probability mass function (log PMF) for a Poisson distribution.

    # Parameters
    - `x`: Where to evaluate the log PMF.
    - `lam`: The rate parameter (lambda), representing the average number of events.

    # Returns
    The log PMF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    return x * lax.log(lam) - lam - jsp.gammaln(x + 1)


def cdf(
    x: Integer[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The cumulative density function (CDF) for a Poisson distribution (P(X <= x)).

    # Parameters
    - `x`: Where to evaluate the CDF.
    - `lam`: The rate parameter (lambda).

    # Returns
    The CDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = jsp.gammainc(x + 1.0, lam)
    result = lax.select(x < 0, jnp.array(0.0), result)

    return result


def logcdf(
    x: Integer[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the cumulative density function (log CDF) for a Poisson distribution.

    # Parameters
    - `x`: Where to evaluate the log CDF (ln P(X <= x)).
    - `lam`: The rate parameter (lambda).

    # Returns
    The log CDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = lax.log(jsp.gammainc(x + 1.0, lam))

    # Handle values outside of support
    result = lax.select(x < 0, -jnp.inf, result)

    return result


def ccdf(
    x: Integer[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The complementary cumulative density function (cCDF) for a Poisson distribution (P(X > x)).

    # Parameters
    - `x`: Where to evaluate the cCDF.
    - `lam`: The rate parameter (lambda).

    # Returns
    The cCDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = jsp.gammaincc(x + 1.0, lam)

    # Handle values outside of support
    result = lax.select(x < 0, jnp.array(1.0), result)

    return result


def logccdf(
    x: Integer[ArrayLike, "..."],
    lam: Real[ArrayLike, "..."],
) -> Real[Array, "..."]:
    """
    The log of the complementary cumulative density function (log cCDF) for a Poisson distribution.

    # Parameters
    - `x`: Where to evaluate the log cCDF.
    - `lam`: The rate parameter (lambda).

    # Returns
    The log cCDF evaluated at `x`. The output will have the broadcasted shapes of `x` and `lam`.
    """
    # Cast to Array
    x, lam = jnp.asarray(x), jnp.asarray(lam)

    result = lax.log(jsp.gammaincc(x + 1.0, lam))

    # Handle values outside of support
    result = lax.select(x < 0, jnp.array(0.0), result)

    return result


class Poisson(Distribution):
    """
    A Poisson distribution.

    # Attributes
    - `lam`: The rate parameter (lambda), representing the average number of events in an interval.
    """

    lam: Node[Real[Array, "..."]]


    def __init__(
        self,
        lam: Real[ArrayLike, "..."] | Node[Real[Array, "..."]],
    ):
        # Initialize rate parameter (lambda)
        if isinstance(lam, Node):
            if isinstance(lam.obj, ArrayLike):
                self.lam = lam # type: ignore
        else:
            self.lam = Observed(jnp.asarray(lam))


    def logprob(self, node: Node) -> Scalar:
        obj, lam = (
            node.obj,
            self.lam.obj,
        )

        # Filter out irrelevant values
        obj, _ = eqx.partition(obj, node._filter_spec)

        # Helper function for the single-leaf log-probability evaluation
        def leaf_logprob(x: Integer[ArrayLike, "..."]) -> Scalar:
            return logprob(x, lam).sum()

        # Compute log probabilities across leaves
        eval_obj = jt.map(leaf_logprob, obj)

        # Compute total sum
        total = jt.reduce_associative(lambda x,y: x + y, eval_obj, identity=0.0)

        return jnp.asarray(total)

    def sample(self, shape: int | Tuple[int, ...], key: PRNGKeyArray = jr.key(0)):
        # Coerce to tuple
        if isinstance(shape, int):
            shape = (shape, )

        return jr.poisson(key, self.lam.obj, shape=shape)



from typing import Tuple

import jax.numpy as jnp
from jaxtyping import PyTree, Scalar

from bayinx.core.constraint import Constraint
from bayinx.core.types import T


class Identity(Constraint):
    """
    Does nothing.
    """

    def constrain(self, obj: T, filter_spec: PyTree) -> Tuple[T, Scalar]:
        """
        Applies the identity transformation (does nothing) and computes its log-Jacobian adjustment (0).

        # Returns
        A tuple containing:
            - The same `PyTree`.
            - A scalar `Array` containing 0.
        """
        log_jac: Scalar = jnp.array(0.0)

        return obj, log_jac



from typing import Callable, Dict, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
from jax.lax import scan
from jaxtyping import Array, Float, PRNGKeyArray, Scalar

from bayinx.core.flow import FlowLayer, FlowSpec


def _h(x: Array) -> Array:
    """Non-linearity for Sylvester flow."""
    return jnp.tanh(x)


def _dh(x: Array) -> Array:
    """Derivative of the non-linearity."""
    return 1.0 - jnp.tanh(x)**2


def _construct_orthogonal(vectors: Float[Array, " dim rank"]) -> Float[Array, " dim rank"]:
    """
    Constructs a D x M orthogonal matrix (Q) using Householder reflections.
    """
    D, M = vectors.shape

    # Initialize Q as the thin identity matrix (first M columns of I_D)
    Q_initial = jnp.eye(D, M)

    def apply_reflection(Q_current, v_k):
        v_norm_sq = jnp.sum(v_k**2)
        tau = 2.0 / v_norm_sq

        a_T = jnp.dot(v_k, Q_current)

        update = jnp.outer(v_k, tau * a_T)

        Q_next = Q_current - update

        return Q_next, None

    Q_final, _ = scan(
        f=apply_reflection,
        init=Q_initial,
        xs=vectors.T
    )

    return Q_final


class SylvesterLayer(FlowLayer):
    """
    A Sylvester flow.

    # Attributes
    - `params`: Dictionary containing raw parameters.
    - `constraints`: Dictionary of constraining transformations.
    - `static`: Whether the flow layer is frozen.
    - `rank`: The rank (M) of the transformation.
    """

    rank: int
    params: Dict[str, Array]
    constraints: Dict[str, Callable[[Array], Array]]
    static: bool

    def __init__(self, dim: int, rank: int, key: PRNGKeyArray = jr.key(0)):
        """
        Initializes the Sylvester flow.

        # Parameters
        - `dim`: The dimension of the parameter space (D).
        - `rank`: The number of hidden units/reflections (M).
        """
        self.static = False
        self.rank = rank

        k1, k2, k3, k4 = jr.split(key, 4)

        # Initialize parameters
        # hvecs: D x M matrix where each column is a Householder vector
        self.params = {
            "hvecs": jr.normal(k1, (dim, rank)) / dim**0.5,
            "r1": jr.normal(k2, (rank, rank)) / rank**0.5,
            "r2": jr.normal(k3, (rank, rank)) / rank**0.5,
            "b": jr.normal(k4, (rank,)) / rank**0.5,
        }

        # Constraint for Upper Triangular matrices with positive diagonal
        def constrain_triangular(matrix: Array):
            # Extract diagonal and apply exponential to ensure invertibility
            diag: Array = jnp.exp(jnp.diag(matrix))
            # Return upper triangular matrix with modified diagonal
            return jnp.fill_diagonal(jnp.triu(matrix), diag, inplace=False)

        self.constraints = {
            "r1": constrain_triangular,
            "r2": constrain_triangular
        }

    def transform_params(self) -> Dict[str, Array]:
        """
        Applies constraints and constructs the orthogonal matrix Q.
        """
        constrained = self.constrain_params()

        # Construct orthogonal matrix
        q = _construct_orthogonal(constrained["hvecs"])
        constrained["q"] = q

        return constrained

    def __forward(self, draw: Float[Array, " n_dim"]) -> Float[Array, " n_dim"]:
        params = self.transform_params()

        # Extract parameters
        Q: Float[Array, "dim rank"] = params["q"]
        R1: Float[Array, "rank rank"] = params["r1"]
        R2: Float[Array, "rank rank"] = params["r2"]
        b: Float[Array, " rank"] = params["b"]

        # Compute inner terms
        y = R2.dot(Q.T.dot(draw)) + b
        h_y = _h(y)

        # Compute forward transform
        draw = draw + Q.dot(R1.dot(h_y))

        return draw

    @eqx.filter_jit
    def forward(self, draws: Float[Array, "n_draws n_dim"]) -> Float[Array, "n_draws n_dim"]:
        f = jax.vmap(self.__forward, 0)
        return f(draws)

    def __adjust(self, draw: Float[Array, " n_dim"]) -> Scalar:
        params = self.transform_params()

        # Extract parameters
        Q: Float[Array, "dim rank"] = params["q"]
        R1: Float[Array, "rank rank"] = params["r1"]
        R2: Float[Array, "rank rank"] = params["r2"]
        b: Float[Array, " rank"] = params["b"]

        # Recompute the argument to the nonlinearity
        term = R2.dot(Q.T.dot(draw)) + b
        diag_h_prime = _dh(term)

        # Diagonal of R1 and R2
        d_r1 = jnp.diag(R1)
        d_r2 = jnp.diag(R2)

        # Diagonal term of the matrix (I + diag(h') R2 R1)
        # diag_term_i = 1 + h'_i * (R2_ii * R1_ii)
        diag_term = 1.0 + diag_h_prime * d_r2 * d_r1

        lja = jnp.sum(jnp.log(diag_term))

        assert lja.shape == ()

        return lja

    @eqx.filter_jit
    def adjust(self, draws: Float[Array, "n_draws n_dim"]) -> Float[Array, "n_draws n_dim"]:
        f = jax.vmap(self.__adjust, 0)
        return f(draws)

    def __forward_and_adjust(self, draw: Float[Array, " n_dim"]) -> Tuple[Float[Array, " n_dim"], Scalar]:
        params = self.transform_params()

        Q: Float[Array, "dim rank"] = params["q"]
        R1: Float[Array, "rank rank"] = params["r1"]
        R2: Float[Array, "rank rank"] = params["r2"]
        b: Float[Array, " rank"] = params["b"]

        # --- Forward ---
        q_z = jnp.dot(Q.T, draw)
        arg = jnp.dot(R2, q_z) + b

        # h(arg)
        h_y = _h(arg)

        # Update
        term = jnp.dot(R1, h_y)
        draw_new = draw + jnp.dot(Q, term)

        # --- Adjust ---
        # Derivative h'(arg)
        diag_h_prime = _dh(arg)

        # Diagonals of triangular matrices
        d_r1 = jnp.diag(R1)
        d_r2 = jnp.diag(R2)

        # Log-det of triangular matrix
        diag_term = 1.0 + diag_h_prime * d_r2 * d_r1
        lja = jnp.sum(jnp.log(jnp.abs(diag_term)))

        assert len(draw_new.shape) == 1
        assert lja.shape == ()

        return draw_new, lja

    @eqx.filter_jit
    def forward_and_adjust(self, draws: Float[Array, "n_draws n_dim"]) -> Tuple[Float[Array, "n_draws n_dim"], Scalar]:
        f = jax.vmap(self.__forward_and_adjust, 0)
        return f(draws)


class Sylvester(FlowSpec):
    """
    A specification for the Orthogonal Sylvester flow.
    """
    rank: int

    def __init__(self, rank: int):
        self.rank = rank

    def construct(self, dim: int) -> SylvesterLayer:
        if self.rank > dim:
            raise ValueError(f"Rank {self.rank} cannot be greater than dimension {dim}.")

        return SylvesterLayer(dim, self.rank)



import threading
from contextlib import contextmanager
from dataclasses import dataclass

import jax.numpy as jnp
from jaxtyping import Scalar

# Local storage for model context
_model_context = threading.local()


@dataclass
class Target:
    value: Scalar

    # Define relevant methods to get around lack of explicit mutability in JAX
    def __iadd__(self, other):
        self.value = self.value + other
        return self

    def __add__(self, other):
        return self.value + other

    def __radd__(self, other):
        return self.value + other


@contextmanager
def model_context():
    """
    Context manager that sets up the implicit `target` accumulator.

    This context allows the `<<` operator for Node classes to automatically
    accumulate log probabilities without requiring explicit handling of target.
    """
    _model_context.target = Target(jnp.array(0.0))

    try:
        yield _model_context.target
    finally: # Remove old context if present
        if hasattr(_model_context, "target"):
            delattr(_model_context, "target")



from typing import Optional

import equinox as eqx
import jax.tree as jt
from jaxtyping import PyTree

from bayinx.core.node import Node
from bayinx.core.types import T


class Observed(Node[T]):
    """
    A container for observed nodes of a probabilistic model.


    # Attributes
    - `obj`: An internal realization of the node.
    - `_filter_spec`: An internal filter specification for `obj`.
    """

    def __init__(
        self, obj: T, filter_spec: Optional[PyTree] = None
    ):
        if filter_spec is None: # Default filter specification
            # Generate empty specification
            filter_spec = jt.map(lambda _: False, obj)

            # Specify array-like leaves
            filter_spec = eqx.tree_at(
                where=lambda obj: obj,
                pytree=filter_spec,
                replace=jt.map(eqx.is_array_like, obj),
            )

        self.obj = obj
        self._filter_spec = filter_spec



from typing import Generic, Self

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jax.flatten_util import ravel_pytree
from jaxtyping import Array, PRNGKeyArray, Scalar

from bayinx.core.variational import M, Variational
from bayinx.dists import normal


class MeanField(Variational, Generic[M]):
    """
    A fully factorized Gaussian approximation to a posterior distribution.

    # Attributes
    - `dim`: The dimension of the support.
    - `mean`: The mean of the unconstrained approximation.
    - `log_std` The log-transformed standard deviation of the unconstrained approximation.
    """

    mean: Array
    log_std: Array

    def __init__(self, model: M):
        """
        Constructs an unoptimized meanfield posterior approximation.

        # Parameters
        - `model`: A probabilistic `Model` object.
        - `init_log_std`: The initial log-transformed standard deviation of the Gaussian approximation.
        """
        # Partition model
        params, self._static = eqx.partition(model, model.filter_spec)

        # Flatten params component
        params, self._unflatten = ravel_pytree(params)

        # Initialize variational parameters
        self.mean = params
        self.log_std = jnp.full(params.size, 0.0)

    @property
    def filter_spec(self):
        # Generate empty specification
        filter_spec = jtu.tree_map(lambda _: False, self)

        # Specify variational parameters
        filter_spec = eqx.tree_at(
            lambda mf: mf.mean,
            filter_spec,
            replace=True,
        )
        filter_spec = eqx.tree_at(
            lambda mf: mf.log_std,
            filter_spec,
            replace=True,
        )

        return filter_spec

    @eqx.filter_jit
    def sample(self, n: int, key: PRNGKeyArray = jr.PRNGKey(0)) -> Array:
        # Sample variational draws
        draws: Array = (
            jr.normal(key=key, shape=(n, self.mean.size)) * jnp.exp(self.log_std)
            + self.mean
        )

        return draws

    @eqx.filter_jit
    def eval(self, draws: Array) -> Array:
        return normal.logprob(
            x=draws,
            mu=self.mean,
            sigma=jnp.exp(self.log_std),
        ).sum(axis=1)

    @eqx.filter_jit
    def elbo(self, n: int, batch_size: int, key: PRNGKeyArray) -> Scalar:
        dyn, static = eqx.partition(self, self.filter_spec)

        @eqx.filter_jit
        def elbo(dyn: Self, n: int, key: PRNGKeyArray) -> Scalar:
            vari = eqx.combine(dyn, static)

            # Sample draws from variational distribution
            draws: Array = vari.sample(n, key)

            # Evaluate posterior density for each draw
            posterior_evals: Array = vari.eval_model(draws)

            # Evaluate variational density for each draw
            variational_evals: Array = vari.eval(draws)

            # Evaluate ELBO
            return jnp.mean(posterior_evals - variational_evals)

        return elbo(dyn, n, key)

    @eqx.filter_jit
    def elbo_grad(self, n: int, batch_size: int, key: PRNGKeyArray) -> Self:
        dyn, static = eqx.partition(self, self.filter_spec)

        @eqx.filter_jit
        @eqx.filter_grad
        def elbo_grad(dyn: Self, n: int, key: PRNGKeyArray):
            vari = eqx.combine(dyn, static)

            # Sample draws from variational distribution
            draws: Array = vari.sample(n, key)

            # Evaluate posterior density for each draw
            posterior_evals: Array = vari.eval_model(draws)

            # Evaluate variational density for each draw
            variational_evals: Array = vari.eval(draws)

            # Evaluate ELBO
            return jnp.mean(posterior_evals - variational_evals)

        return elbo_grad(dyn, n, key)



from typing import List

import jax.numpy as jnp
from jaxtyping import Array

from bayinx import Model, define
from bayinx.dists import Normal
from bayinx.nodes import Continuous, Observed


class MyModel(Model, init=False):
    x: Continuous[Array] = define(shape = (2,2), init = jnp.ones((2,2)))
    y: Observed[Array] = define(shape = (2,2), init = jnp.ones((2,2)))
    z: Observed[List[Array]] = define(init = [jnp.array(-1.0), jnp.array(0.0), jnp.array(1.0)])


    def model(self, target):
        add_node = self.x + self.y
        sub_node = self.x - self.y
        mul_node = self.x * self.y
        div_node = self.x / self.y
        matmul_node = self.x @ self.y


        add_node << Normal(0.0, 1.0)
        sub_node << Normal(0.0, 1.0)
        mul_node << Normal(0.0, 1.0)
        div_node << Normal(0.0, 1.0)
        matmul_node << Normal(0.0, 1.0)

        for node in self.z:
            node << Normal(0.0, 1.0)

        return target


def test_dunder():
    # Construct model
    model = MyModel()

    assert not jnp.isnan(model())



from typing import Generic, Protocol, TypeVar, runtime_checkable

from jaxtyping import PyTree

from bayinx.core.constraint import Constraint

T = TypeVar("T", bound=PyTree)

@runtime_checkable
class HasConstraint(Protocol, Generic[T]):
    """
    Protocol for probabilistic nodes that have constraints.
    """

    obj: T
    _filter_spec: PyTree
    _constraint: Constraint
